# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Tests for cubicweb-datacat schema."""


from cubicweb import ValidationError
from cubicweb.devtools.testlib import CubicWebTC

from utils import create_file, mediatypes_scheme


class DatasetRelationsTC(CubicWebTC):
    """Test case for Dataset entity type relations."""

    def test_dataset_dcat_theme_relation(self):
        """Check that constraint for ``theme`` relation on Dataset entity type is enforced."""
        with self.admin_access.repo_cnx() as cnx:
            # A concept scheme with a concept, and we link a catalog to this scheme
            scheme = cnx.create_entity('ConceptScheme', title=u'My Thesaurus')
            label = cnx.create_entity('Label', label=u'My Concept', kind=u'preferred',
                                      label_of=cnx.create_entity('Concept', in_scheme=scheme))
            concept = label.label_of[0]
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog', description=u'A catalog',
                                    theme_taxonomy=scheme,
                                    catalog_publisher=cnx.create_entity('Agent', name=u'bob'))
            cnx.commit()
            cnx.create_entity('Dataset', title=u'My Dataset', description=u'Dataset 1',
                              in_catalog=cat, dcat_theme=concept)
            cnx.commit()
            # Another scheme not related to the catalog
            scheme2 = cnx.create_entity('ConceptScheme', title=u'My Second Thesaurus')
            label2 = cnx.create_entity('Label', label=u'My Second Concept', kind=u'preferred',
                                       label_of=cnx.create_entity('Concept', in_scheme=scheme2))
            concept2 = label2.label_of[0]
            cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                cnx.create_entity('Dataset', title=u'My Second Dataset', description=u'Dataset 2',
                                  in_catalog=cat, dcat_theme=concept2)
                cnx.commit()
                self.assertEqual(cm.exception.msg,
                                 "Theme must belong to the dataset's catalog vocabulary")
            # To have a theme, a dataset must be in a catalog
            with self.assertRaises(ValidationError) as cm:
                # No in_catalog
                cnx.create_entity('Dataset', title=u'My Third Dataset', description=u'Dataset 3',
                                  dcat_theme=concept)
                cnx.commit()
                self.assertEqual(cm.exception.msg,
                                 "Theme must belong to the dataset's catalog vocabulary")


class ResourceFeedTC(CubicWebTC):

    def test_resourcefeed_read_permissions(self):
        with self.admin_access.cnx() as cnx:
            self.create_user(cnx, u'bob', ('users', ))
            agent = cnx.create_entity('Agent', name=u'doe')
            catalog = cnx.create_entity('DataCatalog', title=u'c',
                                        description=u'd',
                                        catalog_publisher=agent)
            ds = cnx.create_entity('Dataset', title=u'd', description=u'd',
                                   in_catalog=catalog)
            mediatype, = mediatypes_scheme(cnx, u'whatever')
            cnx.create_entity('ResourceFeed',
                              url=u'ftp://user:secret@example.org',
                              media_type=mediatype,
                              resource_feed_of=ds)
            cnx.commit()
        with self.new_access(u'anon').cnx() as cnx:
            rset = cnx.find('ResourceFeed')
            self.assertEqual(len(rset), 0)
        with self.new_access(u'bob').cnx() as cnx:
            rset = cnx.find('ResourceFeed')
            self.assertEqual(len(rset), 0)

    def test_script_mimetype_constraints(self):
        """Check MIME type constraints on validation/transformation_script
        relations.
        """
        with self.admin_access.repo_cnx() as cnx:
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog', description=u'A catalog',
                                    catalog_publisher=cnx.create_entity('Agent', name=u'Publisher'))
            ds = cnx.create_entity('Dataset', title=u'ds', description=u'A dataset', in_catalog=cat)
            allpurpose, textcsv, textplain = mediatypes_scheme(
                cnx, u'application/octet-stream', u'text/csv', u'text/plain')
            cnx.create_entity('ValidationScript', name=u'csv',
                              media_type=textcsv,
                              implemented_by=create_file(cnx, 'pass'))
            cnx.create_entity('ValidationScript', name=u'plain',
                              media_type=textplain,
                              implemented_by=create_file(cnx, 'pass'))
            cnx.create_entity('ValidationScript', name=u'allpurpose',
                              media_type=allpurpose,
                              implemented_by=create_file(cnx, 'pass'))
            cnx.create_entity('ResourceFeed', url=u'a/b/c',
                              media_type=allpurpose,
                              resource_feed_of=ds)
            cnx.commit()
            with self.assertRaises(ValidationError) as cm:
                cnx.execute('SET R validation_script S WHERE S name "csv"')
                cnx.commit()
            self.assertEqual(cm.exception.errors,
                             {'validation_script-subject': (
                                 u'script does not handle resource feed data format')})
            cnx.rollback()
            cnx.execute('SET R validation_script S WHERE S name "allpurpose"')
            cnx.commit()


class ConcepConstraintTC(CubicWebTC):
    """Test case for schema constraints about the ``Concept`` entity type."""

    def test_license_type_constraint(self):
        """Check that a license type can only be set on a concept which is in the right concept
        scheme."""
        with self.admin_access.repo_cnx() as cnx:
            type_scheme = cnx.create_entity('ConceptScheme', title=u'Type scheme')
            type_concept = cnx.create_entity('Concept', in_scheme=type_scheme)
            cnx.create_entity('Label', label_of=type_concept, label=u'Share-alike',
                              kind=u'preferred', language_code=u'en')
            lic_scheme = cnx.create_entity('ConceptScheme', title=u'License scheme')
            other_scheme = cnx.create_entity('ConceptScheme', title=u'Other scheme')
            cnx.execute('SET CS scheme_relation RT WHERE CS eid %(cs)s, RT name %(rt)s',
                        {'cs': lic_scheme.eid, 'rt': 'license_type'})
            cnx.commit()
            # License type can be set on concept in License scheme
            license = cnx.create_entity('Concept', in_scheme=lic_scheme,
                                        license_type=type_concept.eid)
            cnx.create_entity('Label', label_of=license, label=u'My Custom License',
                              kind=u'preferred', language_code=u'en')
            cnx.commit()
            # License type cannot be set on concept in Other scheme
            with self.assertRaises(ValidationError) as cm:
                license = cnx.create_entity('Concept', in_scheme=other_scheme,
                                            license_type=type_concept.eid)
                cnx.create_entity('Label', label_of=license, label=u'My Custom License',
                                  kind=u'preferred', language_code=u'en')
                cnx.commit()
            self.assertEqual(cm.exception.errors,
                             {u'license_type-subject': u'Only concepts in licenses scheme can '
                              'have a license type'})


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat tests for views"""

from datetime import datetime
import json
from tempfile import NamedTemporaryFile

from mock import patch
from pytz import utc
from rdflib.compare import to_isomorphic, graph_diff

from cubicweb.devtools.testlib import CubicWebTC

from cubes.skos.rdfio import default_graph
from cubes.datacat.views import dataset_publisher_choices

from utils import mock_entity_cw_url, mock_concept_cwuri


class PublisherChoicesFunctionTC(CubicWebTC):
    """Test case for cubicweb-datacat dataset publisher choices function."""

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            pub1 = cnx.create_entity('Agent', name=u'Publisher 1')
            pub2 = cnx.create_entity('Agent', name=u'Publisher 2')
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog',
                                    description=u'A nice catalog', catalog_publisher=pub1)
            cnx.commit()
        self.cat_eid = cat.eid
        self.pub1_eid = pub1.eid
        self.pub2_eid = pub2.eid

    def test_choices_add_dataset_from_catalog(self):
        """Check that publisher list is correct when dataset is created from a catalog."""
        with self.admin_access.web_request() as req:
            req.form['__linkto'] = 'in_catalog:{0}:subject'.format(self.cat_eid)
            dset_class = self.vreg['etypes'].etype_class('Dataset')(req)
            add_form = self.vreg['forms'].select('edition', req, entity=dset_class)
            publisher_field = add_form.field_by_name('dataset_publisher', 'subject')
            self.assertEqual(dataset_publisher_choices(add_form, publisher_field),
                             [(u'Publisher 1', unicode(self.pub1_eid))])

    def test_choices_add_dataset_no_catalog(self):
        """Check that publisher list is correct when dataset is created standalone."""
        with self.admin_access.web_request() as req:
            dset_class = self.vreg['etypes'].etype_class('Dataset')(req)
            add_form = self.vreg['forms'].select('edition', req, entity=dset_class)
            publisher_field = add_form.field_by_name('dataset_publisher', 'subject')
            self.assertEqual(dataset_publisher_choices(add_form, publisher_field),
                             [(u'Publisher 1', unicode(self.pub1_eid)),
                              (u'Publisher 2', unicode(self.pub2_eid))])


def sorted_triples(graph):
    return sorted(graph.serialize(format='nt').splitlines())


class RDFAdapterTC(CubicWebTC):
    """Test case for RDF data export."""

    def setup_database(self):
        # patch now() so that computed datetime attributes always return the same datetime
        # See https://docs.python.org/3/library/unittest.mock-examples.html#partial-mocking
        with patch('cubes.datacat.hooks.datetime') as mock_dt:
            mock_dt.now.return_value = datetime(2016, 02, 02, 15, 25, 0, tzinfo=utc)
            mock_dt.side_effect = datetime
            # Now set up entities
            with self.admin_access.repo_cnx() as cnx:
                lic_scheme = cnx.create_entity('ConceptScheme',
                                               cwuri=u'http://publications.europa.eu/'
                                               'resource/authority/licence')
                cnx.execute('SET CS scheme_relation RT WHERE CS eid %(cs)s, RT name %(rt)s',
                            {'cs': lic_scheme.eid, 'rt': 'license_type'})
                scheme = cnx.create_entity('ConceptScheme', cwuri=u'http://example.org/scheme',
                                           title=u'Concept Scheme')
                nat_concept = scheme.add_concept(u'National authority')
                attribution_concept = scheme.add_concept(u'Attribution')
                annual_concept = scheme.add_concept(u'Annual')
                csv_concept = scheme.add_concept(u'CSV')
                xls_concept = scheme.add_concept(u'Excel XLS')
                appxls_concept = scheme.add_concept(u'application/vnd.ms-excel')
                zip_concept = scheme.add_concept(u'ZIP')
                appzip_concept = scheme.add_concept(u'application/zip')
                eng_concept = scheme.add_concept(u'English')
                edu_concept = scheme.add_concept(u'Education, culture and sport')
                publisher = cnx.create_entity('Agent', name=u'The Publisher',
                                              publisher_type=nat_concept,
                                              email=u'publisher@example.org')
                contact = cnx.create_entity('Agent', name=u'The Contact Point',
                                            email=u'contact@example.org')
                license = cnx.create_entity('Concept', in_scheme=lic_scheme,
                                            license_type=attribution_concept)
                cnx.create_entity('Label', label_of=license, kind=u"preferred",
                                  label=u'Other (attribution)')
                cat = cnx.create_entity('DataCatalog', title=u'My Catalog',
                                        description=u'A nice catalog', catalog_publisher=publisher,
                                        homepage=u'http://cat.example.org', language=eng_concept,
                                        theme_taxonomy=scheme, license=license,
                                        issued=datetime(2016, 02, 01, 20, 40, 0, tzinfo=utc),
                                        modified=datetime(2016, 02, 02, 18, 25, 0, tzinfo=utc))
                ds = cnx.create_entity('Dataset', title=u'First Dataset', description=u'A dataset',
                                       in_catalog=cat, dataset_publisher=publisher,
                                       dataset_contact_point=contact, keyword=u'keyword',
                                       dataset_frequency=annual_concept, dcat_theme=edu_concept)
                csv_dist = cnx.create_entity('Distribution', title=u'First Dataset (CSV)',
                                             description=u'First Dataset in CSV format',
                                             of_dataset=ds, license=license,
                                             distribution_format=csv_concept,
                                             access_url=u'http://www.example.org')
                xls_dist = cnx.create_entity('Distribution', title=u'First Dataset (XLS)',
                                             description=u'First Dataset in XLS format',
                                             of_dataset=ds, license=license,
                                             distribution_format=xls_concept,
                                             media_type=appxls_concept,
                                             access_url=u'http://www.example.org')
                zip_dist = cnx.create_entity('Distribution', title=u'First Dataset (ZIP)',
                                             description=u'First Dataset in ZIP format',
                                             of_dataset=ds, license=license,
                                             distribution_format=zip_concept,
                                             media_type=appzip_concept,
                                             access_url=u'http://www.example.org')
                cnx.commit()
                self.entities_to_mock = [
                    (scheme.eid, 'http://example.org/concept_scheme'),
                    (publisher.eid, 'http://example.org/publisher'),
                    (contact.eid, 'http://example.org/contact'),
                    (cat.eid, u'http://example.org/catalog'),
                    (ds.eid, 'http://example.org/dataset'),
                    (csv_dist.eid, 'http://example.org/dataset/csv'),
                    (xls_dist.eid, 'http://example.org/dataset/xls'),
                    (zip_dist.eid, 'http://example.org/dataset/zip'),
                ]
                self.concepts_to_mock = [
                    (nat_concept.eid, u'http://example.org/national_authority'),
                    (attribution_concept.eid, u'http://example.org/attribution'),
                    (annual_concept.eid, u'http://example.org/annual'),
                    (csv_concept.eid, u'http://example.org/csv'),
                    (xls_concept.eid, u'http://example.org/xls'),
                    (zip_concept.eid, u'http://example.org/zip'),
                    (appxls_concept.eid, u'http://example.org/application_xls'),
                    (appzip_concept.eid, u'http://example.org/application_zip'),
                    (eng_concept.eid, u'http://example.org/english'),
                    (edu_concept.eid, u'http://example.org/education'),
                    (license.eid, 'http://example.org/license'),
                ]
                self.cat_eid = cat.eid

    @patch('cubes.skos.rdfio.RDFGraphGenerator.same_as_uris')
    def _check_rdf_view(self, action_regid, expected_fname, mock_same_as_uris):
        # Because we mock either cwuri or absolute_url, RDFGraphGenerator will add `sameAS` triples
        # on exported entities. We don't want this
        mock_same_as_uris.return_value = []
        with self.admin_access.client_cnx() as cnx:
            # Mock entities to have predictable URIs (no eid)
            for eid, fake_url in self.entities_to_mock:
                mock_entity_cw_url(cnx, eid, fake_url)
            for eid, fake_url in self.concepts_to_mock:
                mock_concept_cwuri(cnx, eid, fake_url)
            # Now call export
            cat = cnx.entity_from_eid(self.cat_eid)
            with NamedTemporaryFile() as f:
                f.write(cat.view(action_regid))
                f.seek(0)
                graph = default_graph()
                graph.load('file://' + f.name, rdf_format='xml')
            expected_graph = default_graph()
            expected_graph.load('file://' + self.datapath(expected_fname), rdf_format='xml')
            iso1 = to_isomorphic(graph._graph)
            iso2 = to_isomorphic(expected_graph._graph)
            in_both, in_first, in_second = graph_diff(iso1, iso2)
            self.assertEqual(iso1, iso2, u'RDF graphs are not the same:\n'
                             '* Only in actual:\n{0}\n\n'
                             '* Only in expected:\n{1}'.format(
                                 sorted_triples(in_first), sorted_triples(in_second)))

    def test_complete_rdf_view(self):
        """Check that 'RDF export' action produce valid DCAT RDF data."""
        self._check_rdf_view('dcat.rdf.complete', 'valid_export.xml')

    def test_ckan_rdf_view(self):
        """Check that 'RDF export for CKAN' action produce RDF data suitable for CKAN."""
        self._check_rdf_view('dcat.rdf.ckan', 'ckan_export.rdf')


class CkanJsonLicenseExportTC(CubicWebTC):
    """Test case for JSON license export to be used with CKAN."""

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            lic_scheme = cnx.create_entity('ConceptScheme',
                                           cwuri=u'http://publications.europa.eu/'
                                           'resource/authority/licence')
            cnx.execute('SET CS scheme_relation RT WHERE CS eid %(cs)s, RT name %(rt)s',
                        {'cs': lic_scheme.eid, 'rt': 'license_type'})
            # Licence with match in OpenDefinition by URL
            cc_by_40 = cnx.create_entity('Concept', in_scheme=lic_scheme,
                                         cwuri=u'http://creativecommons.org/licenses/by/4.0/')
            cnx.create_entity('Label', label_of=cc_by_40,
                              label=u'Creative Commons Attribution 4.0 International',
                              kind=u"preferred", language_code=u'en')
            # Licence with match in OpenDefinition by title
            pddl = cnx.create_entity('Concept', in_scheme=lic_scheme,
                                     cwuri=u'http://opendatacommons.org/licenses/pddl/1.0/')
            cnx.create_entity('Label', label_of=pddl,
                              label=u'Open Data Commons Public Domain Dedication and License v1.0',
                              kind=u"preferred", language_code=u'en')
            # Licence with no match in OpenDefinition
            other_at = cnx.create_entity('Concept', in_scheme=lic_scheme,
                                         cwuri=u'http://example.org/lo-ol')
            cnx.create_entity('Label', label_of=other_at, label=u'Licence Ouverte / Open Licence',
                              kind=u"preferred", language_code=u'en')
            cnx.commit()
            self.lic_scheme_eid = lic_scheme.eid

    def test_ckan_json_license_scheme_export(self):
        """Check that a license concept scheme is exported correctly for CKAN."""
        with self.admin_access.web_request() as req:
            lic_scheme = req.execute('ConceptScheme CS WHERE CS scheme_relation RT, '
                                     'RT name "license_type"').one()
            export = json.loads(lic_scheme.view('dcat.ckan.json.licenses'))
            with open(self.datapath('licenses.json')) as f:
                expected = json.load(f)
            self.assertEqual(export, expected)


if __name__ == '__main__':
    import unittest
    unittest.main()



"""cubicweb-datacat test utilities"""

from mock import Mock

from cubicweb import Binary


def create_file(cnx, data, data_name=None, **kwargs):
    """Create a File entity"""
    data_name = data_name or data.decode('utf-8')
    kwargs.setdefault('data_format', u'text/plain')
    return cnx.create_entity('File', data=Binary(data),
                             data_name=data_name,
                             **kwargs)


def produce_file(cnx, resourcefeed, inputfile):
    """Simulate the production of `inputfile` by resource feed `resourcefeed`"""
    tseq = resourcefeed.transformation_script[0].transformation_sequence
    # Build a transformation process "by hand".
    with cnx.security_enabled(write=False):
        process = cnx.create_entity(
            'DataTransformationProcess',
            process_input_file=inputfile,
            process_for_resourcefeed=resourcefeed,
            transformation_sequence=tseq)
        cnx.commit()  # transformation triggered in a hook.
    return process.reverse_produced_by[0]


def mock_entity_cw_url(cnx, eid, returned_url):
    """Mock an entity's ``absolute_url()`` method by returning a given and fixed URL.

    Entity to mock is given by its eid.
    """
    entity = cnx.entity_from_eid(eid)
    entity.absolute_url = Mock(return_value=returned_url)


def mock_concept_cwuri(cnx, concept_eid, returned_url):
    """Mock a ``Concept`` entity's ``cwuri`` attribute by returning a given and fixed URL.

    Concept to mock is given by its eid.
    """
    concept = cnx.entity_from_eid(concept_eid)
    concept_mock = Mock(return_value=returned_url)

    # rdflib check if returned URI contains invalid chars; always say it does not
    def contains(self, item):
        return False

    concept_mock.__contains__ = contains
    # rdflib uses repr as string serialization
    concept_mock.__repr__ = Mock(return_value=returned_url)

    concept.absolute_url = concept_mock
    concept.cw_attr_cache['cwuri'] = concept_mock


def mediatypes_scheme(cnx, *labels):
    """Build a "dummy" media-types concept scheme to satisfy hooks on
    distribution/file media type conformance. Return concepts corresponding to
    requested `labels`.
    """
    scheme = cnx.create_entity(
        'ConceptScheme',
        cwuri=u'http://www.iana.org/assignments/media-types/media-types.xml')
    return [scheme.add_concept(label) for label in labels]



# copyright 2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-datacat command plugins"""

from __future__ import print_function

import io
import json
import os
import os.path as osp
import requests
import zipfile

from logilab.common.shellutils import ProgressBar

from cubicweb.cwctl import CWCTL
from cubicweb.dataimport.importer import SimpleImportLog
from cubicweb.toolsutils import Command, underline_title
from cubicweb.utils import admincnx

from cubes.datacat import cwsource_pull_data
from cubes.datacat.vocabularies import (ADMS_FRENCH_LABELS, EUROVOC_URI, LICENSE_SCHEME,
                                        eurovoc_domains, simplify_eurovoc)
from cubes.skos import rdfio
from cubes.skos.ccplugin import ImportSkosData
from cubes.skos.sobjects import graph_extentities, import_skos_extentities


class SyncSKOSSchemes(Command):
    """Synchronize SKOS schemes for a datacat instance.

    <instance>
      identifier of a datacat instance.

    """
    arguments = '<instance>'
    name = 'sync-schemes'
    min_args = 1
    max_args = 1

    def run(self, args):
        appid = args[0]
        with admincnx(appid) as cnx:
            rset = cnx.execute(
                'Any S ORDERBY X WHERE X is SKOSSource, X through_cw_source S')
            title = '-> synchronizing SKOS sources'
            pb = ProgressBar(len(rset), title=title)
            created, updated = set([]), set([])
            for eid, in rset:
                stats = cwsource_pull_data(cnx.repo, eid, raise_on_error=False)
                pb.update()
                created.update(stats['created'])
                updated.update(stats['updated'])
            # French translations
            qs = (u'Any C WHERE C cwuri IN ({0}), '
                  u'NOT EXISTS(L label_of C, L kind "preferred", L language_code "fr")'.format(
                      ', '.join([u"'{0}'".format(uri) for uri in ADMS_FRENCH_LABELS])))
            uri2eid = dict((concept.cwuri, concept.eid) for concept in cnx.execute(qs).entities())
            title = '-> Adding missing french labels for ADMS concepts'
            pb = ProgressBar(len(uri2eid), title=title)
            for concept_uri, concept_eid in uri2eid.items():
                fr_label = ADMS_FRENCH_LABELS.get(concept_uri)
                if fr_label:
                    label = cnx.create_entity('Label', label=fr_label, language_code=u'fr',
                                              kind=u'preferred', label_of=concept_eid)
                    created.add(label)
                pb.update()
            cnx.commit()
            print('\n   {0} created, {1} updated'.format(len(created), len(updated)))
            print('\n-> Adding constraints for license schemes')
            cnx.execute('SET CS scheme_relation RT WHERE CS cwuri %(scheme_uri)s, '
                        'RT name %(rtype_name)s',
                        {'scheme_uri': LICENSE_SCHEME, 'rtype_name': 'license_type'})
            cnx.commit()


def set_store_default(cls):
    """Set default value for "cw-store" option to "massive"."""
    for option in cls.options:
        if option[0] == 'cw-store':
            option[1]['default'] = 'massive'
            break
    return cls


@set_store_default
class ImportEurovoc(ImportSkosData):
    """Synchronize Eurovoc concept schemes for a datacat instance.

    <instance>
      identifier of a datacat instance.

    """
    arguments = '[options] <instance>'
    name = 'import-eurovoc'
    min_args = 1
    options = (ImportSkosData.options + (
        ('update', {
            'short': 'u', 'type': 'yn', 'default': False,
            'help': 'Force Eurovoc thesaurus download, even file already exists'
        }),
    ))

    def run(self, args):
        appid = args[0]
        update = self.get('update')
        with admincnx(appid) as cnx:
            datadir = osp.join(cnx.repo.config.appdatahome, 'eurovoc')
        if not osp.isdir(datadir):
            os.mkdir(datadir)
        zip_fname = osp.join(datadir, 'eurovoc_skos.zip')
        n3_fname = osp.join(datadir, 'eurovoc_skos.n3')
        simplified_fname = osp.join(datadir, 'eurovoc_skos_simple.n3')
        json_fname = osp.join(datadir, 'eurovoc_domains.json')
        print(u'\n{0}'.format(underline_title('Importing Eurovoc thesaurus')))
        if not osp.isfile(zip_fname) or update:
            print(u'Downloading eurovoc SKOS/RDF ZIP file from publications.europa.eu')
            r = requests.get(EUROVOC_URI, stream=True)
            with open(zip_fname, 'wb') as f:
                for chunk in r.iter_content(chunk_size=8192):
                    if chunk:
                        f.write(chunk)
        elif osp.isfile(zip_fname):
            print(u'Using existing eurovoc SKOS/RDF ZIP file {0}'.format(zip_fname))
        if not osp.isfile(n3_fname) or update:  # n3 file is faster to load
            print(u'Extracting RDF file and converting to Turtle')
            eurovoc_graph = rdfio.default_graph()
            with zipfile.ZipFile(zip_fname) as eurovoc_zip:
                eurovoc_graph.load(eurovoc_zip.open('eurovoc_skos.rdf'), rdf_format='xml')
                with open(n3_fname, 'w') as f:
                    eurovoc_graph.dump(f, rdf_format='n3')
        print(u'Preparing data for import')
        simplified_graph = rdfio.default_graph()
        if not osp.isfile(simplified_fname) or update:
            eurovoc_graph = rdfio.default_graph()
            with open(n3_fname) as f:
                eurovoc_graph.load(f, rdf_format='n3')
            # Flatten and simplify the graph
            simplify_eurovoc(eurovoc_graph, simplified_graph)
            with open(simplified_fname, 'w') as f:
                simplified_graph.dump(f, rdf_format='n3')
        else:
            with open(simplified_fname) as f:
                simplified_graph.load(f, rdf_format='n3')
        if not osp.isfile(json_fname) or update:
            with io.open(json_fname, 'w', encoding='utf-8') as f:
                f.write(unicode(json.dumps(eurovoc_domains(simplified_graph), ensure_ascii=False,
                                           sort_keys=True, indent=2, separators=(u',', u':'))))
            print(u'-> JSON file with Eurovoc domains created at {0}'.format(json_fname))
        print(u'Importing data')
        with admincnx(appid) as cnx:
            import_log = SimpleImportLog('eurovoc.import')
            cwsource = cnx.execute(u'Any S WHERE S is CWSource, S url %(eurovoc_uri)s',
                                   {'eurovoc_uri': EUROVOC_URI}).one()
            store = self.cw_store_factories[self.get('cw-store')](cnx)
            (created, updated), conceptschemes = import_skos_extentities(
                cnx, graph_extentities(simplified_graph), import_log, source=cwsource, store=store)
            cnx.commit()
            print(u'Created: {0}\nUpdated: {0}'.format(len(created), len(updated)))
            if conceptschemes:
                print(u'Concept schemes:\n* {0}'.format('\n* '.join(conceptschemes)))
            _update_source(cnx, cwsource)
            cnx.repo.shutdown()


def _update_source(cnx, cwsource):
    """Update cw_source relation of entities to point to Eurovoc CWSource.
    """
    sql = cnx.system_sql
    sqlqs = ("UPDATE cw_source_relation SET eid_to=%(source)s WHERE eid_from in ("
             "SELECT _X.cw_eid FROM cw_{0} AS _X WHERE _X.cw_cwuri ILIKE '%%eurovoc%%')")
    for tablename in ('Label', 'Concept', 'ConceptScheme'):
        sql(sqlqs.format(tablename), {'source': cwsource.eid})
    cnx.commit()


for cmd in (SyncSKOSSchemes, ImportEurovoc):
    CWCTL.register(cmd)



"""cubicweb-datacat application package

Data catalog
"""


def register_catalog_rdf_mapping(reg):
    """Register mapping between DCAT RDF vocabulary and DataCatalog entity type."""
    reg.register_prefix('dcat', 'http://www.w3.org/ns/dcat#')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_etype_equivalence('DataCatalog', 'dcat:Catalog')
    reg.register_attribute_equivalence('DataCatalog', 'title', 'dcterms:title')
    reg.register_attribute_equivalence('DataCatalog', 'description', 'dcterms:description')
    reg.register_attribute_equivalence('DataCatalog', 'issued', 'dcterms:issued')
    reg.register_attribute_equivalence('DataCatalog', 'modified', 'dcterms:modified')
    reg.register_relation_equivalence('DataCatalog', 'catalog_publisher', 'Agent',
                                      'dcterms:publisher')
    reg.register_relation_equivalence('Dataset', 'in_catalog', 'DataCatalog', 'dcat:dataset',
                                      reverse=True)


def register_dataset_rdf_mapping(reg):
    """Register mapping between DCAT RDF vocabulary and Dataset entity type."""
    reg.register_prefix('dcat', 'http://www.w3.org/ns/dcat#')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_etype_equivalence('Dataset', 'dcat:Dataset')
    reg.register_attribute_equivalence('Dataset', 'title', 'dcterms:title')
    reg.register_attribute_equivalence('Dataset', 'description', 'dcterms:description')
    reg.register_attribute_equivalence('Dataset', 'issued', 'dcterms:issued')
    reg.register_attribute_equivalence('Dataset', 'modified', 'dcterms:modified')
    reg.register_attribute_equivalence('Dataset', 'identifier', 'dcterms:identifier')
    reg.register_relation_equivalence('Dataset', 'in_catalog', 'DataCatalog', 'dcat:dataset',
                                      reverse=True)
    reg.register_relation_equivalence('Dataset', 'dataset_publisher', 'Agent',
                                      'dcterms:publisher')
    reg.register_relation_equivalence('Dataset', 'dataset_contact_point', 'Agent',
                                      'dcat:contactPoint')
    reg.register_relation_equivalence('Distribution', 'of_dataset', 'Dataset', 'dcat:distribution',
                                      reverse=True)


def register_distribution_rdf_mapping(reg):
    """Register mapping between DCAT RDF vocabulary and Distribution entity type."""
    reg.register_prefix('dcat', 'http://www.w3.org/ns/dcat#')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_etype_equivalence('Distribution', 'dcat:Distribution')
    reg.register_attribute_equivalence('Distribution', 'title', 'dcterms:title')
    reg.register_attribute_equivalence('Distribution', 'description', 'dcterms:description')
    reg.register_attribute_equivalence('Distribution', 'issued', 'dcterms:issued')
    reg.register_attribute_equivalence('Distribution', 'modified', 'dcterms:modified')
    reg.register_relation_equivalence('Distribution', 'of_dataset', 'Dataset', 'dcat:distribution',
                                      reverse=True)


def register_publisher_rdf_mapping(reg):
    """Register mapping between FOAF RDF vocabulary and Agent entity type."""
    reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
    reg.register_prefix('dcterms', 'http://purl.org/dc/terms/')
    reg.register_etype_equivalence('Agent', 'foaf:Agent')
    reg.register_attribute_equivalence('Agent', 'name', 'foaf:name')
    reg.register_relation_equivalence('Dataset', 'dataset_publisher', 'Agent',
                                      'dcterms:publisher')


def register_contact_point_rdf_mapping(reg):
    """Register mapping between VCard RDF vocabulary and Agent entity type."""
    reg.register_prefix('vcard', 'http://www.w3.org/2006/vcard/ns#')
    reg.register_prefix('dcat', 'http://www.w3.org/ns/dcat#')
    reg.register_relation_equivalence('Dataset', 'dataset_contact_point', 'Agent',
                                      'dcat:contactPoint')
    reg.register_etype_equivalence('Agent', 'vcard:Kind')
    reg.register_attribute_equivalence('Agent', 'name', 'vcard:fn')


def concept_in(cnx, label, scheme_uri):
    """Return a concept with the given label and in the given concept scheme."""
    rset = cnx.execute('Any C WHERE C in_scheme S, S cwuri %(scheme_uri)s, '
                       'L label_of C, L kind "preferred", L label %(label)s',
                       {'label': label, 'scheme_uri': scheme_uri})
    if rset:
        return rset.one()
    else:
        return None


def cwsource_pull_data(repo, cwsource_eid, **kwargs):
    """Pull data from a CWSource associated with a ResourceFeed"""
    dfsource = repo.sources_by_eid[cwsource_eid]
    kwargs.setdefault('force', True)
    kwargs.setdefault('raise_on_error', True)
    with repo.internal_cnx() as icnx:
        stats = dfsource.pull_data(icnx, **kwargs)
        icnx.commit()
    return stats



"""cubicweb-datacat unit tests for entities"""

import datetime

from cubicweb.devtools.testlib import CubicWebTC

from cubes.skos import rdfio


class RDFAdapterTC(CubicWebTC):
    """Test case for RDF data export."""

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            lic_scheme = cnx.create_entity('ConceptScheme', cwuri=u'http://publications.europa.eu/'
                                           'resource/authority/licence')
            cnx.execute('SET CS scheme_relation RT WHERE CS eid %(cs)s, RT name %(rt)s',
                        {'cs': lic_scheme.eid, 'rt': 'license_type'})
            scheme = cnx.create_entity('ConceptScheme', cwuri=u'http://example.org/scheme')
            nat_concept = scheme.add_concept(u'National authority')
            attribution_concept = scheme.add_concept(u'Attribution')
            annual_concept = scheme.add_concept(u'Annual')
            csv_concept = scheme.add_concept(u'CSV')
            xls_concept = scheme.add_concept(u'Excel XLS')
            appxls_concept = scheme.add_concept(u'application/vnd.ms-excel')
            zip_concept = scheme.add_concept(u'ZIP')
            appzip_concept = scheme.add_concept(u'application/zip')
            eng_concept = scheme.add_concept(u'English')
            edu_concept = scheme.add_concept(u'Education, culture and sport')
            publisher = cnx.create_entity('Agent', name=u'The Publisher',
                                          publisher_type=nat_concept,
                                          email=u'publisher@example.org')
            contact = cnx.create_entity('Agent', name=u'The Contact Point',
                                        email=u'contact@example.org')
            license = cnx.create_entity('Concept', in_scheme=lic_scheme,
                                        cwuri=u'http://creativecommons.org/licenses/by/3.0',
                                        license_type=attribution_concept)
            cnx.create_entity('Label', label_of=license, kind=u"preferred",
                              label=u'Creative Commons Attribution')
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog',
                                    description=u'A nice catalog', catalog_publisher=publisher,
                                    homepage=u'http://cat.example.org', language=eng_concept,
                                    theme_taxonomy=scheme, license=license,
                                    issued=datetime.datetime(2016, 02, 01, 20, 40, 00),
                                    modified=datetime.datetime(2016, 02, 02, 18, 25, 00))
            ds = cnx.create_entity('Dataset', title=u'First Dataset', description=u'A nice datacat',
                                   in_catalog=cat, dataset_publisher=publisher,
                                   dataset_contact_point=contact, keyword=u'keyword',
                                   dataset_frequency=annual_concept, dcat_theme=edu_concept)
            dist1 = cnx.create_entity('Distribution', title=u'First Dataset (CSV)',
                                      description=u'First Dataset in CSV format', of_dataset=ds,
                                      license=license, distribution_format=csv_concept,
                                      access_url=u'http://www.example.org')
            dist2 = cnx.create_entity('Distribution', title=u'First Dataset (XLS)',
                                      description=u'First Dataset in XLS format', of_dataset=ds,
                                      license=license, distribution_format=xls_concept,
                                      media_type=appxls_concept,
                                      access_url=u'http://www.example.org')
            dist3 = cnx.create_entity('Distribution', title=u'First Dataset (ZIP)',
                                      description=u'First Dataset in ZIP format', of_dataset=ds,
                                      license=license, distribution_format=zip_concept,
                                      media_type=appzip_concept,
                                      access_url=u'http://www.example.org')
            cnx.commit()
            self.cat_eid = cat.eid
            self.ds_eid = ds.eid
            self.dist1_eid = dist1.eid
            self.dist2_eid = dist2.eid
            self.dist3_eid = dist3.eid
            self.publisher_eid = publisher.eid
            self.contact_eid = contact.eid
            self.license_eid = license.eid
            self.nat_concept_uri = nat_concept.cwuri
            self.attribution_concept_uri = attribution_concept.cwuri
            self.annual_concept_uri = annual_concept.cwuri
            self.csv_concept_uri = csv_concept.cwuri
            self.xls_concept_uri = xls_concept.cwuri
            self.appxls_concept_uri = appxls_concept.cwuri
            self.zip_concept_uri = zip_concept.cwuri
            self.appzip_concept_uri = appzip_concept.cwuri
            self.eng_concept_uri = eng_concept.cwuri
            self.theme_scheme_uri = scheme.cwuri
            self.edu_concept_uri = edu_concept.cwuri

    def test_rdf_export_catalog(self):
        """Check that we get expected RDF data when exporting a catalog."""
        with self.admin_access.repo_cnx() as cnx:
            cat = cnx.entity_from_eid(self.cat_eid)
            publisher = cnx.entity_from_eid(self.publisher_eid)
            cat_uri = cat.absolute_url()
            rdfcat = cat.cw_adapt_to('RDFPrimary')
            graph = rdfio.default_graph()
            rdfcat.fill(graph)
            self._check_literal_property(graph,
                                         cat_uri, 'http://purl.org/dc/terms/title', u'My Catalog')
            self._check_uri_property(graph, cat_uri, 'http://purl.org/dc/terms/publisher',
                                     publisher.absolute_url())
            self._check_uri_property(graph, cat_uri, 'http://xmlns.com/foaf/0.1/homepage',
                                     'http://cat.example.org')
            self._check_uri_property(graph, cat_uri, 'http://purl.org/dc/terms/language',
                                     self.eng_concept_uri)
            self._check_uri_property(graph, cat_uri, 'http://www.w3.org/ns/dcat#themeTaxonomy',
                                     self.theme_scheme_uri)
            self._check_literal_property(graph, publisher.absolute_url(),
                                         'http://xmlns.com/foaf/0.1/name',
                                         u'The Publisher')

    def test_rdf_export_dataset(self):
        """Check that we get expected RDF data when exporting a catalog."""
        with self.admin_access.repo_cnx() as cnx:
            ds = cnx.entity_from_eid(self.ds_eid)
            publisher = cnx.entity_from_eid(self.publisher_eid)
            contact = cnx.entity_from_eid(self.contact_eid)
            ds_uri = ds.absolute_url()
            rdfds = ds.cw_adapt_to('RDFPrimary')
            graph = rdfio.default_graph()
            rdfds.fill(graph)
            self._check_literal_property(graph,
                                         ds_uri, 'http://purl.org/dc/terms/title', u'First Dataset')
            self._check_literal_property(graph,
                                         ds_uri, 'http://www.w3.org/ns/dcat#keyword',
                                         u'keyword')
            self._check_uri_property(graph, ds_uri, 'http://purl.org/dc/terms/publisher',
                                     publisher.absolute_url())
            self._check_uri_property(graph, ds_uri, 'http://www.w3.org/ns/dcat#contactPoint',
                                     contact.absolute_url())
            self._check_uri_property(graph, ds_uri, 'http://purl.org/dc/terms/accrualPeriodicity',
                                     self.annual_concept_uri)
            self._check_uri_property(graph, ds_uri, 'http://www.w3.org/ns/dcat#theme',
                                     self.edu_concept_uri)

    def test_rdf_export_distribution(self):
        """Check that we get expected RDF data when exporting a distribution."""
        with self.admin_access.repo_cnx() as cnx:
            for dist_eid, title, filetype, mediatype in [
                (self.dist1_eid, u'First Dataset (CSV)', self.csv_concept_uri, None),
                (self.dist2_eid, u'First Dataset (XLS)', self.xls_concept_uri,
                 self.appxls_concept_uri),
                (self.dist3_eid, u'First Dataset (ZIP)', self.zip_concept_uri,
                 self.appzip_concept_uri),
            ]:
                dist = cnx.entity_from_eid(dist_eid)
                license = cnx.entity_from_eid(self.license_eid)
                dist_uri = dist.absolute_url()
                rdfdist = dist.cw_adapt_to('RDFPrimary')
                graph = rdfio.default_graph()
                rdfdist.fill(graph)
                self._check_literal_property(graph, dist_uri, 'http://purl.org/dc/terms/title',
                                             title)
                if title == u'First Dataset (CSV)':
                    self._check_uri_property(graph, dist_uri, 'http://purl.org/dc/terms/license',
                                             license.cwuri)
                    self._check_uri_property(graph, license.cwuri,
                                             'http://purl.org/dc/terms/type',
                                             self.attribution_concept_uri)
                    self._check_uri_property(graph, license.cwuri,
                                             'http://www.w3.org/1999/02/22-rdf-syntax-ns#type',
                                             'http://purl.org/dc/terms/LicenseDocument')
                    self._check_literal_property(graph, license.cwuri,
                                                 'http://purl.org/dc/terms/title',
                                                 u'Creative Commons Attribution')
                if filetype is not None:
                    self._check_uri_property(graph, dist_uri, 'http://purl.org/dc/terms/format',
                                             filetype)
                if mediatype is not None:
                    self._check_uri_property(graph, dist_uri, 'http://www.w3.org/ns/dcat#mediaType',
                                             mediatype)

    def test_rdf_export_publisher(self):
        """Check that we get expected RDF data when exporting a publisher."""
        with self.admin_access.repo_cnx() as cnx:
            publisher = cnx.entity_from_eid(self.publisher_eid)
            publisher_uri = publisher.absolute_url()
            rdfpub = publisher.cw_adapt_to('RDFPrimary')
            graph = rdfio.default_graph()
            rdfpub.fill(graph)
            self._check_literal_property(graph, publisher_uri, 'http://xmlns.com/foaf/0.1/name',
                                         u'The Publisher')
            self._check_literal_property(graph, publisher_uri, 'http://xmlns.com/foaf/0.1/mbox',
                                         u'mailto:publisher@example.org')
            self._check_uri_property(graph, publisher_uri, 'http://purl.org/dc/terms/type',
                                     self.nat_concept_uri)

    def test_rdf_export_contact_point(self):
        """Check that we get expected RDF data when exporting a contact point."""
        with self.admin_access.repo_cnx() as cnx:
            contact = cnx.entity_from_eid(self.contact_eid)
            contact_uri = contact.absolute_url()
            rdfcontact = contact.cw_adapt_to('RDFContactPoint')
            graph = rdfio.default_graph()
            rdfcontact.fill(graph)
            self._check_literal_property(graph, contact_uri, 'http://www.w3.org/2006/vcard/ns#fn',
                                         u'The Contact Point')
            self._check_literal_property(graph, contact_uri,
                                         'http://www.w3.org/2006/vcard/ns#hasEmail',
                                         u'mailto:contact@example.org')

    def _check_literal_property(self, graph, subject_uri, rdf_property, expected_value):
        """Check that, in the given graph, `subject_uri` has the expected value for the
        `dcterms:title` property."""
        self.assertEqual(list(graph.objects(subject_uri, graph.uri(rdf_property))),
                         [expected_value])

    def _check_uri_property(self, graph, subject_uri, rdf_property, expected_uri):
        """Check that, in the given graph, `subject_uri` is related via `rdf_property` to a
        `foaf:Agent` with expected URI."""
        self.assertEqual(list(graph.objects(subject_uri, graph.uri(rdf_property))), [expected_uri])


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2014-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat server objects"""

import hashlib

from cubicweb import Binary
from cubicweb.server.sources import datafeed
from cubicweb.dataimport import NoHookRQLObjectStore, MetaGenerator
from cubicweb.dataimport.importer import (ExtEntity, ExtEntitiesImporter,
                                          HTMLImportLog)


def compute_sha1hex(value):
    """Return the SHA1 hex digest of `value`."""
    return unicode(hashlib.sha1(value).hexdigest())


class ResourceFeedParser(datafeed.DataFeedParser):
    """Fetch files and apply validation and transformation processes before
    attaching them to their Dataset as Resources.
    """
    __regid__ = 'datacat.resourcefeed-parser'

    def importer(self, url, extid2eid):
        """Return an ExtEntity importer."""
        cnx = self._cw
        schema = cnx.vreg.schema
        # XXX Using NoHookRQLObjectStore because RQLObjectStore does not
        # support setting cw_source relation.
        metagenerator = MetaGenerator(cnx, source=self.source)
        store = NoHookRQLObjectStore(cnx, metagenerator)
        import_log = HTMLImportLog(url)
        return ExtEntitiesImporter(schema, store, import_log=import_log,
                                   extid2eid=extid2eid)

    def extid2eid(self):
        qs = 'Any H,X WHERE X data_sha1hex H, X cw_source S, S eid %(s)s'
        args = {'s': self.source.eid}
        return dict(self._cw.execute(qs, args))

    def process(self, url, raise_on_error=False):
        """Build a File entity from data fetched from url"""
        stream = self.retrieve_url(url)
        data = stream.read()
        sha1hex = compute_sha1hex(data)
        extid2eid = self.extid2eid()
        extentity = ExtEntity('File', sha1hex)
        if extentity.extid not in extid2eid:
            # Only set `values` when the entity does not already exist
            # (otherwise the importer will consider the entity for update
            # which is not what we want).
            extentity.values.update(
                {'data_name': set([url.split('/')[-1]]),
                 'data_sha1hex': set([sha1hex]),
                 'data': set([Binary(data)])}
            )
        importer = self.importer(url, extid2eid)
        importer.import_entities([extentity])
        created, updated = importer.created, importer.updated
        self.debug('data import for %s completed: created %d entities, updated %s entities',
                   url, len(created), len(updated))
        entity = None
        if created:
            assert not updated, 'unexpected update of entities {0}'.format(updated)
            assert len(created) == 1, created
            entity = self._cw.entity_from_eid(created.pop())
            self.stats['created'].add(entity.eid)
        if updated:
            assert not created, 'unexpected creation of entities {0}'.format(created)
            assert len(updated) == 1, updated
            entity = self._cw.entity_from_eid(updated.pop())
            self.stats['updated'].add(entity.eid)
        if entity is not None:
            # Ensure imported entity as a cw_source, otherwise re-import would
            # not work.
            assert entity.cw_source
            self.process_data(entity)

    def process_data(self, entity):
        """Launch validation and transformation scripts of all related
        ResourceFeed entities.
        """
        cwsource = self._cw.entity_from_eid(self.source.eid)
        data_format = None
        for resourcefeed in cwsource.reverse_resource_feed_source:
            if data_format:
                # XXX Better do this in a schema constraint on
                # `resource_feed_source` relation.
                if resourcefeed.data_format != data_format:
                    raise ValueError('MIME types of resource feeds attached to '
                                     'CWSource #%d mismatch' % cwsource.eid)
            else:
                data_format = resourcefeed.data_format
                entity.cw_set(data_format=resourcefeed.data_format,
                              file_distribution=resourcefeed.resourcefeed_distribution)
            # Link imported file to distribution.
            self._cw.execute(
                'SET F file_distribution D WHERE F eid %(f)s, R resourcefeed_distribution D, '
                'R eid %(r)s, NOT EXISTS(F file_distribution D)',
                {'f': entity.eid, 'r': resourcefeed.eid})
            # Run the validation script for imported file.
            vprocess = None
            if resourcefeed.validation_script:
                vscript = resourcefeed.validation_script[0]
                vprocess = self._cw.create_entity('DataValidationProcess',
                                                  process_for_resourcefeed=resourcefeed,
                                                  validation_script=vscript,
                                                  process_input_file=entity)
                self.debug('created validation process #%d for file #%d',
                           vprocess.eid, entity.eid)
            if resourcefeed.transformation_script:
                tseq = resourcefeed.transformation_script[0].transformation_sequence
                tprocess = self._cw.create_entity('DataTransformationProcess',
                                                  process_for_resourcefeed=resourcefeed,
                                                  transformation_sequence=tseq,
                                                  process_input_file=entity,
                                                  process_depends_on=vprocess)
                self.debug('created transformation process #%d for file #%d',
                           tprocess.eid, entity.eid)



"""cubicweb-datacat unit tests for hooks"""

from datetime import datetime

from pytz import utc

from cubicweb.devtools.testlib import CubicWebTC

from utils import create_file, produce_file, mediatypes_scheme


def transformation_script(cnx, script_code, mtype):
    return cnx.create_entity('TransformationScript', name=u'script',
                             media_type=mtype,
                             implemented_by=create_file(cnx, script_code))


class ResourceFeedHooksTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog', description=u'A catalog',
                                    catalog_publisher=cnx.create_entity('Agent', name=u'Publisher'))
            ds = cnx.create_entity('Dataset', title=u'Test dataset', description=u'A dataset',
                                   in_catalog=cat)
            cnx.commit()
            self.dataset_eid = ds.eid

    def test_distribution_created(self):
        with self.admin_access.repo_cnx() as cnx:
            mtype, = mediatypes_scheme(cnx, u'text/plain')
            resourcefeed = cnx.create_entity(
                'ResourceFeed', url=u'a/b/c',
                media_type=mtype,
                resource_feed_of=self.dataset_eid)
            cnx.commit()
            self.assertTrue(resourcefeed.resourcefeed_distribution)
            dist = resourcefeed.resourcefeed_distribution[0]
            self.assertEqual(dist.media_type[0].label(), u'text/plain')
            self.assertEqual([x.eid for x in dist.of_dataset],
                             [self.dataset_eid])

    def test_resourcefeed_cwsource(self):
        with self.admin_access.repo_cnx() as cnx:
            mtype, = mediatypes_scheme(cnx, u'whatever')
            resourcefeed = cnx.create_entity(
                'ResourceFeed', url=u'a/b/c',
                media_type=mtype,
                resource_feed_of=self.dataset_eid)
            cnx.commit()
            source = resourcefeed.resource_feed_source[0]
            self.assertEqual(source.url, resourcefeed.url)
            resourcefeed.cw_set(url=u'c/b/a')
            cnx.commit()
            source.cw_clear_all_caches()
            self.assertEqual(source.url, u'c/b/a')
            resourcefeed1 = cnx.create_entity(
                'ResourceFeed', url=u'c/b/a',
                media_type=mtype,
                resource_feed_of=self.dataset_eid)
            cnx.commit()
            self.assertEqual(resourcefeed1.resource_feed_source[0].eid,
                             source.eid)

    def test_resourcefeed_transformation_sequence(self):
        """Test for datacat.add/drop-transformation-sequence hooks."""
        with self.admin_access.cnx() as cnx:
            mtype, = mediatypes_scheme(cnx, u'whatever')
            script = cnx.create_entity(
                'TransformationScript', name=u'script',
                media_type=mtype,
                implemented_by=create_file(cnx, 'pass'))
            resourcefeed = cnx.create_entity(
                'ResourceFeed', url=u'a/b/c',
                media_type=mtype,
                transformation_script=script,
                resource_feed_of=self.dataset_eid)
            cnx.commit()
            tseq = cnx.find('TransformationSequence').one()
            tstep = tseq.reverse_in_sequence[0]
            self.assertEqual(tstep.index, 0)
            self.assertEqual(tstep.step_script, [script])
            resourcefeed.cw_set(transformation_script=None)
            cnx.commit()
            self.assertFalse(cnx.find('TransformationSequence'))

    def test_linkto_dataset(self):
        with self.admin_access.repo_cnx() as cnx:
            inputfile = create_file(cnx, 'data')
            mtype, = mediatypes_scheme(cnx, u'whatever')
            code = (
                'from __future__ import print_function\n'
                'print("plop", end="")\n'
            )
            tscript = transformation_script(cnx, code, mtype)
            resourcefeed = cnx.create_entity('ResourceFeed', url=u'a/b/c',
                                             media_type=mtype,
                                             resource_feed_of=self.dataset_eid,
                                             transformation_script=tscript)
            cnx.commit()
            produce_file(cnx, resourcefeed, inputfile)
            rset = cnx.execute('Any X WHERE X file_distribution D, D eid %s' %
                               resourcefeed.resourcefeed_distribution[0].eid)
            self.assertEqual(len(rset), 1, rset)
            outdata = rset.get_entity(0, 0).read()
            self.assertEqual(outdata, 'plop')

    def test_file_replaced(self):
        with self.admin_access.repo_cnx() as cnx:
            mtype, = mediatypes_scheme(cnx, u'whatever')
            tscript = transformation_script(cnx, 'pass', mtype)
            resourcefeed = cnx.create_entity('ResourceFeed', url=u'a/b/c',
                                             resource_feed_of=self.dataset_eid,
                                             media_type=mtype,
                                             transformation_script=tscript)
            cnx.commit()
            outfile1 = produce_file(cnx, resourcefeed,
                                    create_file(cnx, 'data'))
            outfile2 = produce_file(cnx, resourcefeed,
                                    create_file(cnx, 'data 2'))
            outfile3 = produce_file(cnx, resourcefeed,
                                    create_file(cnx, 'data 3'))
            rset = cnx.execute('Any F1,F2 WHERE F1 replaces F2')
            self.assertEqual(rset.rowcount, 2)
            self.assertIn([outfile2.eid, outfile1.eid], rset.rows)
            self.assertIn([outfile3.eid, outfile2.eid], rset.rows)
            rset = cnx.execute(
                'Any X WHERE X file_distribution D, D eid %(d)s',
                {'d': resourcefeed.resourcefeed_distribution[0].eid})
            self.assertEqual(rset.rowcount, 1)
            self.assertEqual(rset[0][0], outfile3.eid)


class DatesHooksTC(CubicWebTC):

    def test_issued_modified(self):
        past = datetime(2008, 1, 4, 3, 4, 5, tzinfo=utc)
        future = datetime(3000, 1, 4, 3, 4, 5, tzinfo=utc)
        with self.admin_access.cnx() as cnx:
            catalog = cnx.create_entity(
                'DataCatalog', title=u'c', description=u'd', issued=past,
                catalog_publisher=cnx.create_entity('Agent', name=u'publisher'))
            dataset = cnx.create_entity(
                'Dataset', title=u'ds', description=u'A dataset',
                in_catalog=catalog)
            dist = cnx.create_entity(
                'Distribution', title=u'di', of_dataset=dataset,
                issued=future, modified=past)
            cnx.commit()
            self.assertEqual(catalog.issued, past)
            self.assertEqual(catalog.modified, past)
            self.assertGreater(dataset.issued, past)
            self.assertEqual(dataset.modified, dataset.issued)
            self.assertGreater(dataset.issued, past)
            self.assertEqual(dist.modified, past)
            self.assertEqual(dist.issued, future)
            dist.cw_set(access_url=u'http://example.org')
            cnx.commit()
            self.assertGreater(dist.modified, past)
            self.assertEqual(dist.issued, future)


class DistributionHooksTC(CubicWebTC):

    def test_accessurl(self):
        with self.admin_access.cnx() as cnx:
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog', description=u'A catalog',
                                    catalog_publisher=cnx.create_entity('Agent', name=u'Publisher'))
            ds = cnx.create_entity('Dataset', identifier=u'ds', title=u'ds',
                                   description=u'A dataset', in_catalog=cat)
            dist_with_url = cnx.create_entity(
                'Distribution', of_dataset=ds, title=u'with',
                access_url=u'http://example.org')
            dist_without_url = cnx.create_entity(
                'Distribution', of_dataset=ds, title=u'without')
            cnx.commit()
            self.assertEqual(dist_with_url.access_url, u'http://example.org')
            self.assertEqual(dist_without_url.access_url,
                             dist_without_url.absolute_url())


class FileDistributionRelationHooksTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog', description=u'A catalog',
                                    catalog_publisher=cnx.create_entity('Agent', name=u'Publisher'))
            ds = cnx.create_entity('Dataset', identifier=u'ds', title=u'ds',
                                   description=u'A dataset', in_catalog=cat)
            self.distribution_eid = cnx.create_entity(
                'Distribution', of_dataset=ds, title=u'THE distr',
                description=u'one and only', access_url=u'http://example.org').eid
            cnx.commit()

    @staticmethod
    def create_file(cnx, **kwargs):
        """Simplified version of utils.create_file"""
        return create_file(cnx, data_name=u"foo.txt", data="xxx", **kwargs)

    def create_distribution_file(self, cnx, **kwargs):
        """Create a file link to self.distribution_eid"""
        return self.create_file(cnx, file_distribution=self.distribution_eid,
                                **kwargs)

    def test_update_file_of_distribution(self):
        with self.admin_access.repo_cnx() as cnx:
            file_v1 = self.create_distribution_file(cnx)
            cnx.commit()
            distr = cnx.find('Distribution', eid=self.distribution_eid).one()
            self.assertEqual(distr.reverse_file_distribution[0], file_v1)
            file_v2 = self.create_file(cnx, replaces=file_v1)
            cnx.commit()
            distr = cnx.find('Distribution', eid=self.distribution_eid).one()
            self.assertEqual(distr.reverse_file_distribution[0], file_v2)
            self.assertEqual(file_v2.replaces[0], file_v1)

    def test_update_distribution_on_create_filedistribution(self):
        with self.admin_access.repo_cnx() as cnx:
            mediatypes_scheme(cnx, u'text/csv')
            cnx.commit()
            distr = cnx.entity_from_eid(self.distribution_eid)
            self.assertIsNone(distr.byte_size)
            self.assertFalse(distr.distribution_format)
            self.assertFalse(distr.media_type)
            self.assertIsNone(distr.download_url)
            distr_file = self.create_distribution_file(cnx, data_format=u'text/csv')
            cnx.commit()
            self.assertEqual(distr.byte_size, distr_file.size())
            self.assertEqual(distr.media_type[0].label(), distr_file.data_format)
            self.assertEqual(distr.access_url,
                             distr_file.cw_adapt_to("IDownloadable").download_url())
            self.assertIsNone(distr.download_url)

    def test_update_distribution_on_create_and_on_update(self):
        """Test the value of issued and modified
            - when a file is added : issued ~= modified = time when the file is added
            - when a new file is added : issued ~= modified = time when the new file is added
            - when the last file is updated : issued = time when the file was added,
                                              modified = time when the file was updated
            - when an older file is updated : nothing change
        """
        with self.admin_access.repo_cnx() as cnx:
            distr = cnx.entity_from_eid(self.distribution_eid)
            file_v1 = self.create_distributionfile_and_check_date(cnx, distr)
            file_v2 = self.create_distributionfile_and_check_date(cnx, distr)
            self.update_file_and_check_date(cnx, distr, file_v2, current_file=True)
            self.update_file_and_check_date(cnx, distr, file_v1, current_file=False)

    def create_distributionfile_and_check_date(self, cnx, distr):
        before = distr.creation_date
        distr_file = self.create_distribution_file(cnx)
        cnx.commit()
        after = datetime.now(utc)
        distr.cw_clear_all_caches()
        for date in [distr.issued, distr.modified]:
            self.assertGreater(date, before)
            self.assertLess(date, after)
        return distr_file

    def update_file_and_check_date(self, cnx, distr, distr_file, current_file):
        issued_before, modified_before = distr.issued, distr.modified
        distr_file.cw_set(data_name=u'bar.txt')
        cnx.commit()
        distr.cw_clear_all_caches()
        self.assertEqual(distr.issued, issued_before)
        if current_file:
            self.assertGreater(distr.modified, modified_before)
        else:
            self.assertEqual(distr.modified, modified_before)

    def test_set_title_description(self):
        with self.admin_access.repo_cnx() as cnx:
            distr_file = self.create_distribution_file(cnx)
            self.assertEqual(distr_file.title, 'THE distr')
            self.assertEqual(distr_file.description, 'one and only')


if __name__ == '__main__':
    from logilab.common.testlib import unittest_main
    unittest_main()



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat schema"""

from yams.buildobjs import (EntityType, RelationDefinition, String,
                            RichString, TZDatetime, Int)
from cubicweb.schema import (RRQLExpression, ERQLExpression,
                             RQLConstraint,
                             RQLVocabularyConstraint)

from cubes.file.schema import File
from cubes.skos.schema import SKOSSource
from cubes.dataprocessing import schema as dataprocessing


_ = unicode


# Until https://www.cubicweb.org/ticket/12437063 gets done.
SKOSSource.__permissions__ = SKOSSource.__permissions__.copy()
SKOSSource.__permissions__['read'] = ('managers', 'users')


def publication_metadata(cls):
    """Class decorator that adds publication metadata to an entity type."""
    cls.add_relation(TZDatetime(description=_('publication date')),
                     name='issued')
    cls.add_relation(TZDatetime(description=_('modification date')),
                     name='modified')
    return cls


def in_scheme_constraint(scheme_uri):
    return RQLVocabularyConstraint(
        'O in_scheme CS, CS cwuri "{0}"'.format(scheme_uri))


class scheme_relation(RelationDefinition):
    """Special relation from a concept scheme to a relation type, that may be used to restrict
    possible concept of a particular relation without depending on the scheme's name or other weak
    mecanism.
    """
    __permissions__ = {'read': ('managers', 'users', 'guests'),
                       'add': ('managers',),
                       'delete': ('managers',)}
    subject = 'ConceptScheme'
    object = 'CWRType'
    cardinality = '**'
    constraints = [
        RQLVocabularyConstraint('NOT EXISTS(C theme_taxonomy S)'),
        RQLVocabularyConstraint('O name "license_type"'),
    ]


class Agent(EntityType):
    name = String(required=True, fulltextindexed=True)
    email = String()
    description = _(
        'An agent (eg. person, group, software or physical artifact).')


class publisher_type(RelationDefinition):
    subject = 'Agent'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _("type of publisher (if unspecified agent is not a publisher)")
    constraints = [
        in_scheme_constraint('http://purl.org/adms/publishertype/1.0'),
    ]


@publication_metadata
class DataCatalog(EntityType):
    """A data catalog is a curated collection of metadata about datasets."""
    title = String(required=True, fulltextindexed=True)
    description = RichString(required=True, fulltextindexed=True)
    homepage = String(description=_('URL to the catalog home page'))


class language(RelationDefinition):
    subject = 'DataCatalog'
    object = 'Concept'
    description = _('languages used in the catalog')


class theme_taxonomy(RelationDefinition):
    subject = 'DataCatalog'
    object = 'ConceptScheme'
    description = _(
        'vocabulary or thesaurus containing a list of themes for datasets in this catalog')


class catalog_publisher(RelationDefinition):
    subject = 'DataCatalog'
    object = 'Agent'
    inlined = True
    cardinality = '1*'
    description = _(
        'agent responsible for making the catalog available')


@publication_metadata
class Dataset(EntityType):
    # DCAT attributes.
    identifier = String(fulltextindexed=True)
    title = String(required=True, fulltextindexed=True)
    description = RichString(required=True, fulltextindexed=True)
    keyword = String(fulltextindexed=True,
                     description=_(
                         'comma-separated list of keywords or tags describing the dataset'))
    landing_page = String(
        description=_('web page providing access to the dataset, its '
                      'distributions and/or additional information'))
    provenance = String(fulltextindexed=True,  # Not in DCAT-AP spec.
                        description=_('source from which this dataset has been produced or '
                                      'extracted'))


class dataset_frequency(RelationDefinition):
    subject = 'Dataset'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _('frequency at which a dataset is updated')
    constraints = [
        in_scheme_constraint(
            'http://publications.europa.eu/resource/authority/frequency'),
    ]


class in_catalog(RelationDefinition):
    subject = 'Dataset'
    object = 'DataCatalog'
    cardinality = '1*'
    inlined = True
    description = _('catalog to which the dataset belongs')


class of_dataset(RelationDefinition):
    subject = 'Distribution'
    object = 'Dataset'
    cardinality = '1*'
    composite = 'object'
    inlined = True
    description = _('dataset to which the distribution is related')


class dataset_contact_point(RelationDefinition):
    subject = 'Dataset'
    object = 'Agent'
    cardinality = '?*'
    description = _('contact information that can be used for flagging '
                    'errors in the dataset or sending comments')
    constraints = [RQLVocabularyConstraint('NOT EXISTS(O publisher_type T)')]


class dataset_publisher(RelationDefinition):
    subject = 'Dataset'
    object = 'Agent'
    cardinality = '?*'
    description = _('agent responsible for making the dataset available')
    constraints = [RQLVocabularyConstraint('S in_catalog C, C catalog_publisher O')]


class dataset_contributors(RelationDefinition):
    subject = 'Dataset'
    object = 'Agent'
    cardinality = '**'
    description = _("agent(s) contributing to the dataset's production")


class dcat_theme(RelationDefinition):
    subject = 'Dataset'
    object = 'Concept'
    constraints = [RQLConstraint('O in_scheme CS, S in_catalog C, C theme_taxonomy CS',
                                 msg=_("Theme must belong to a dataset's catalog vocabulary"))]


class license_type(RelationDefinition):
    subject = 'Concept'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    constraints = [
        RQLConstraint('S in_scheme CS, CS scheme_relation RT, RT name "license_type"',
                      msg=_('Only concepts in licenses scheme can have a license type')),
        in_scheme_constraint('http://purl.org/adms/licencetype/1.0'),
    ]


class license(RelationDefinition):
    subject = ('DataCatalog', 'Distribution')
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    composite = 'subject'
    description = _('the licence under which the entity is made available')
    constraints = [RQLVocabularyConstraint('O in_scheme CS, CS scheme_relation RT, '
                                           'RT name "license_type"')]


@publication_metadata
class Distribution(EntityType):
    title = String(fulltextindexed=True)
    description = RichString(fulltextindexed=True)
    access_url = String(
        description=_('A URL that gives access to a distribution of the '
                      'dataset. The resource at the access URL may contain '
                      'information about how to get the dataset.')
    )
    download_url = String(
        description=_('URL to directly download the distribution. '
                      '(Usually unspecified if the distribution is associated '
                      'with a file in the catalog.)')
    )
    byte_size = Int()


class distribution_media_type(RelationDefinition):
    name = 'media_type'
    subject = 'Distribution'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _('The media type of the distribution as defined in the official register of '
                    'media types managed by IANA')


class distribution_format(RelationDefinition):
    subject = 'Distribution'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _('The file format of the distribution')


class file_distribution(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': (RRQLExpression('U has_update_permission O'), ),
        'delete': (RRQLExpression('U has_update_permission O, '
                                  'NOT EXISTS(S produced_by R)'), ),
    }
    subject = 'File'
    object = 'Distribution'
    cardinality = '??'
    inlined = True
    composite = 'object'
    constraints = [
        RQLVocabularyConstraint('NOT EXISTS(SC implemented_by S)'),
    ]


class replaces(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': (
            RRQLExpression(
                'U has_update_permission O, NOT EXISTS(O produced_by P)'),
        ),
        'delete': (
            RRQLExpression(
                'U has_update_permission O, U has_update_permission S, '
                'NOT EXISTS(S produced_by P), NOT EXISTS(O produced_by P1)'),
        ),
    }
    subject = 'File'
    object = 'File'
    cardinality = '??'
    inlined = True


class ResourceFeed(EntityType):
    """A resource feed is associated to a dataset distribution. It regularly
    fetches files from the specified URL and attached them to this
    distribution.
    """
    __permissions__ = {
        'read': ('managers', ERQLExpression('X owned_by U')),
        'update': (
            'managers',
            ERQLExpression('X owned_by U,'
                           ' NOT EXISTS(X resourcefeed_distribution D, F file_distribution D)')),
        'delete': ('managers', ERQLExpression('X owned_by U')),
        'add': ('managers', 'users')
    }
    title = String(
        description=_('this title will be used to set the title '
                      'of the associated distribution'))
    url = String(required=True)
    __unique_toguether__ = [('url', 'resource_feed_of')]


class resourcefeed_media_type(RelationDefinition):
    name = 'media_type'
    subject = 'ResourceFeed'
    object = 'Concept'
    cardinality = '1*'
    inlined = True
    description = _('MIME type of files served by the resource feed.')


class resource_feed_of(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', RRQLExpression('U has_update_permission O')),
        'delete': ('managers', RRQLExpression('U has_update_permission O')),
    }
    subject = 'ResourceFeed'
    object = 'Dataset'
    cardinality = '1*'
    composite = 'object'
    inlined = True


class resourcefeed_distribution(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': (),
        'delete': (),
    }
    subject = 'ResourceFeed'
    object = 'Distribution'
    cardinality = '1*'
    inlined = True


class resource_feed_source(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': (),
        'delete': (),
    }
    subject = 'ResourceFeed'
    object = 'CWSource'
    cardinality = '1*'
    inlined = True
    # TODO a constraint checking compatibility of data_format between resource
    # feeds sharing the same CWSource.


# Set File permissions:
#  * use `produced_by` relation to prevent modification of generated files
#  * use relative permissions on the Distribution which the file may be a resource
#    of
#  * bind the update permissions on the _Script which uses the File as
#    implementation if any
# XXX Overrides cubicweb-dataprocessing (but we have no obvious way to
# extend ERQLExpression, so...).
_update_file_perms = ('managers', ERQLExpression(
    'NOT EXISTS(X produced_by Y), '
    'NOT EXISTS(X file_distribution D1)'
    ' OR EXISTS(X file_distribution D, U has_update_permission D), '
    'NOT EXISTS(S1 implemented_by X)'
    ' OR EXISTS(S implemented_by X, U has_update_permission S)'
))
File.__permissions__ = File.__permissions__.copy()
File.__permissions__.update({'update': _update_file_perms,
                             'delete': _update_file_perms})


# Customization of dataprocessing for ResourceFeed

dataprocessing.TransformationScript.remove_relation('output_format')

# TransformationSequence is not exposed, restrict its permissions.
dataprocessing.TransformationSequence.__permissions__ = {
    'read': ('managers', 'users'),
    'add': (),
    'update': (),
    'delete': (),
}


class output_media_type(RelationDefinition):
    subject = 'TransformationScript'
    object = 'Concept'
    cardinality = '?*'
    inlined = True
    description = _('MIME type of data generated by the script (if unspecified, '
                    'the input file format will be used)')


class script_media_type(RelationDefinition):
    name = 'media_type'
    subject = ('ValidationScript', 'TransformationScript')
    object = 'Concept'
    cardinality = '1*'
    inlined = True
    description = _('MIME type of files handled by the script (use '
                    '"application/octet-stream" for a generic script)')


class _resourcefeed_script(RelationDefinition):
    __abstract__ = True
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', 'users'),
        'delete': (RRQLExpression('U has_update_permission S'), )
    }
    subject = 'ResourceFeed'
    cardinality = '?*'
    inlined = True
    constraints = [
        RQLConstraint(
            'O media_type C, '
            '(L label_of C, L label "application/octet-stream") OR S media_type C',
            msg=_('script does not handle resource feed data format')),
    ]


class validation_script(_resourcefeed_script):
    description = u'a Script used to validated files from this resource'
    object = 'ValidationScript'


class transformation_script(_resourcefeed_script):
    description = u'a Script used to transform files from this resource'
    object = 'TransformationScript'


class process_for_resourcefeed(RelationDefinition):
    __permissions__ = {
        'read': ('managers', 'users', 'guests'),
        'add': ('managers', RRQLExpression('U has_update_permission O')),
        'delete': ('managers', RRQLExpression('U has_update_permission O')),
    }
    subject = ('DataTransformationProcess', 'DataValidationProcess')
    object = 'ResourceFeed'
    cardinality = '1*'
    composite = 'object'


dataprocessing.process_depends_on.__permissions__ = {
    'read': ('managers', 'users', 'guests'),
    'add': ('managers', RRQLExpression(
        'NOT EXISTS(S process_for_resourcefeed RF1) '
        'OR (S process_for_resourcefeed RF, U has_update_permission RF)')),
    'delete': ('managers', RRQLExpression(
        'NOT EXISTS(S process_for_resourcefeed RF1) '
        'OR (S process_for_resourcefeed RF, U has_update_permission RF)')),
}


dataprocessing.process_depends_on.constraints = [
    RQLConstraint(
        'NOT EXISTS(S process_for_resourcefeed RF1,'
        '           S process_for_resourcefeed RF2) '
        'OR EXISTS(S process_for_resourcefeed RF,'
        '          O process_for_resourcefeed RF)',
        msg=_('dependency between data process must be within the same '
              'resource feed')),
]


dataprocessing.implemented_by.constraints = [
    RQLVocabularyConstraint(
        'NOT EXISTS(O file_distribution DI, DI of_dataset D)'),
]


dataprocessing.DataTransformationProcess.__permissions__['add'] = ()
dataprocessing.DataTransformationProcess.__permissions__['update'] = ()
dataprocessing.DataValidationProcess.__permissions__['add'] = ()
dataprocessing.DataValidationProcess.__permissions__['update'] = ()



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""Migration instructions for target version 0.4.0."""

sync_schema_props_perms('Agent')

# New dependencies
add_cube('skos')

# New attributes
add_attribute('Distribution', 'byte_size')
add_attribute('Agent', 'dcat_type')
add_attribute('DataCatalog', 'homepage')
add_attribute('DataCatalog', 'license')

# New relations
add_relation_type('replaces')
add_relation_type('theme_taxonomy')
add_relation_type('dcat_theme')
add_relation_type('catalog_publisher')

sync_schema_props_perms('file_distribution')

# A distribution is now related to at most one file: the most recent one.
# So for each distribution, keep only the last file_distribution relation, drop the other ones, and
# and add replaces relation between files.
for dist_eid, in rql('Any D WHERE D is Distribution, EXISTS(F file_distribution D)'):
    rset = rql('Any F ORDERBY DATE DESC WHERE F file_distribution D, F creation_date DATE, '
               'D eid %(dist_eid)s', {'dist_eid': dist_eid})
    if len(rset) <= 1:
        continue
    newf_eid = rset.rows[0][0]
    for oldf_eid, in rset.rows[1:]:
        rql('DELETE F file_distribution D WHERE F eid %(f)s, D eid %(d)s',
            {'f': oldf_eid, 'd': dist_eid})
        rql('SET NEWF replaces OLDF WHERE NEWF eid %(n)s, OLDF eid %(o)s',
            {'n': newf_eid, 'o': oldf_eid})
        commit(ask_confirm=False)
        newf_eid = oldf_eid

add_cube('squareui')

add_relation_definition('File', 'produced_by', 'DataTransformationProcess')
add_relation_definition('File', 'validated_by', 'DataValidationProcess')
rql('SET F produced_by P WHERE P process_script S, F produced_by S')
commit()
rql('SET F validated_by P WHERE P process_script S, F validated_by S')
commit()
drop_relation_definition('File', 'produced_by', 'Script')
drop_relation_definition('File', 'validated_by', 'Script')
drop_relation_type('produced_from')
sync_schema_props_perms('validated_by')
sync_schema_props_perms('file_distribution')
sync_schema_props_perms('File')

sync_schema_props_perms('Script')
sync_schema_props_perms('process_input_file')
sync_schema_props_perms('validation_script')
sync_schema_props_perms('transformation_script')

for stype in ('validation', 'transformation'):
    for rf in rql('Any R,RF,SF WHERE R data_format RF, S accepted_format SF, NOT S accepted_format RF, '
                  'R {0}_script S'.format(stype)).entities():
        print('WARNING: conflicting format between resource feed '
              'and its {0} script: {0} / {0}'.format(stype))

commit()



# copyright 2014-2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-datacat views/forms/actions/components for web ui"""

from functools import partial
from copy import deepcopy
from os.path import dirname, join
import json
import re
import urlparse

from logilab.mtconverter import xml_escape

from cubicweb import tags
from cubicweb.predicates import (adaptable, has_related_entities, is_instance,
                                 one_line_rset, score_entity)
from cubicweb.uilib import js
from cubicweb.utils import json_dumps
from cubicweb.web import action, component, facet as cwfacet, formwidgets as fw
from cubicweb.web.views import (actions, ajaxcontroller, baseviews, ibreadcrumbs, json as jsonviews,
                                navigation, uicfg)
from cubicweb.view import EntityView

from cubes.relationwidget import views as rwdg
from cubes.skos import rdfio
from cubes.skos.views import rdf as rdfviews

_ = unicode

abaa = uicfg.actionbox_appearsin_addmenu
afs = uicfg.autoform_section
affk = uicfg.autoform_field_kwargs
pvdc = uicfg.primaryview_display_ctrl
pvs = uicfg.primaryview_section


uicfg.indexview_etype_section['TransformationSequence'] = 'subobject'
# already shown in "administration" section
uicfg.indexview_etype_section['SKOSSource'] = 'subobject'


def build_selectable_uicfg(rtag, select):
    """Return a copy of uicfg `rtag` with `select` set."""
    copy = deepcopy(rtag)
    copy.__module__ = __name__
    copy.__select__ = select
    return copy


def _autocomplete_widget(fname, wdgclass=fw.LazyRestrictedAutoCompletionWidget,
                         placeholder=None):
    if placeholder is None:
        placeholder = _("Start to type your choice: you'll be presented a list of "
                        "matching terms")
    return wdgclass(
        attrs={'placeholder': placeholder, 'size': 82},
        autocomplete_initfunc=fname,
        autocomplete_settings={'limit': 100, 'delay': 300},
    )


def search_in_scheme(req, scheme_uri):
    """Query ConceptScheme with `scheme_uri` for a Label matching `q` term
    found in request's form.
    """
    term = req.form['q']
    limit = req.form.get('limit', 50)
    lang = req.lang
    qs = ('DISTINCT Any X LIMIT {0} '
          'WHERE X in_scheme S, S cwuri %(scheme)s, L label_of X, L label LL,'
          '      L label ILIKE %(term)s'.format(limit))
    args = {'scheme': scheme_uri, 'term': u'%%%s%%' % term}
    return [{'value': concept.eid, 'label': concept.label(language_code=lang)}
            for concept in req.execute(qs, args).entities()]


class NoRedirectAddRelatedActions(actions.AddRelatedActions):
    """Override 'addrelated' action to avoid redirection to the parent entity."""

    __select__ = (actions.AddRelatedActions.__select__ &
                  is_instance('Agent', 'DataCatalog', 'Dataset', 'Distribution'))

    def redirect_params(self, *args, **kwargs):
        return {}


class WorkflowableEntityOutOfContextView(baseviews.OutOfContextView):
    """Out of context view showing workflow state"""

    __select__ = baseviews.OutOfContextView.__select__ & adaptable('IWorkflowable')

    def cell_call(self, row, col, **kwargs):
        self.w(u'<div style="margin-top: 1px;" class="clearfix">')
        super(WorkflowableEntityOutOfContextView, self).cell_call(row, col, **kwargs)
        entity = self.cw_rset.get_entity(row, col)
        iwf = entity.cw_adapt_to('IWorkflowable')
        self.w(tags.span(iwf.printable_state, klass='badge pull-right'))
        self.w(u'</div>')


#
# RDF views
#

class DatasetDCATRDFView(rdfviews.RDFPrimaryView):
    """RDF view for Dataset entities, outputting complete information about
    catalog, distributions and agents.

    This is mainly useful for clients who *do not follow URIs* to gather information about related
    resources and expect all information at one URL (eg. CKAN).
    """

    __regid__ = 'dcat.rdf.complete'
    __select__ = rdfviews.RDFPrimaryView.__select__ & is_instance('Dataset')

    def entity_call(self, entity, graph=None):
        # Copied from parent class
        dump = graph is None
        if graph is None:
            graph = rdfio.default_graph()
        rdfgenerator = entity.cw_adapt_to(self.adapter)
        rdfgenerator.fill(graph)
        # Also output related entities
        for rtype, role, adapter in (('in_catalog', 'subject', 'RDFPrimary'),
                                     ('of_dataset', 'object', 'RDFPrimary'),
                                     ('dataset_publisher', 'subject', 'RDFPrimary'),
                                     ('dataset_contact_point', 'subject', 'RDFContactPoint')):
            for related in entity.related(rtype, role, entities=True):
                rdfgenerator = related.cw_adapt_to(adapter)
                rdfgenerator.fill(graph)
        # Copied from parent class
        if dump:
            self._dump(graph)


class DataCatalogDCATRDFView(rdfviews.RDFPrimaryView):
    """RDF view outputting complete information about catalog and children entities.

    This is mainly useful for clients who *do not follow URIs* to gather information about related
    resources and expect all information at one URL (eg. CKAN).
    """

    __regid__ = 'dcat.rdf.complete'
    __select__ = (rdfviews.RDFPrimaryView.__select__ &
                  has_related_entities('in_catalog', role='object'))

    def entity_call(self, entity, graph=None):
        # Copied from parent class
        dump = graph is None
        if graph is None:
            graph = rdfio.default_graph()
        for dataset in entity.related('in_catalog', role='object').entities():
            dataset.view(self.__regid__, graph=graph)
        # Copied from parent class
        if dump:
            self._dump(graph)


class DataCatalogCkanDCATRDFView(rdfviews.RDFPrimaryView):
    """RDF view outputting complete information about catalog and children entities, suitable for
    CKAN.

    This is mainly useful for clients who *do not follow URIs* to gather information about related
    resources and expect all information at one URL (eg. CKAN).
    """

    __regid__ = 'dcat.rdf.ckan'
    __select__ = (rdfviews.RDFPrimaryView.__select__ &
                  has_related_entities('in_catalog', role='object'))
    adapter = 'RDFCkan'

    def entity_call(self, entity, graph=None):
        # Copied from parent class
        dump = graph is None
        if graph is None:
            graph = rdfio.default_graph()
        for dataset in entity.related('in_catalog', role='object').entities():
            dataset.view(self.__regid__, graph=graph)
        # Copied from parent class
        if dump:
            self._dump(graph)


class DatasetCkanDCATRDFView(rdfviews.RDFPrimaryView):
    """RDF view for Dataset entities, suitable for CKAN, outputting complete information about
    catalog, distributions and agents.

    This is mainly useful for clients who *do not follow URIs* to gather information about related
    resources and expect all information at one URL (eg. CKAN).
    """

    __regid__ = 'dcat.rdf.ckan'
    __select__ = rdfviews.RDFPrimaryView.__select__ & is_instance('Dataset')
    adapter = 'RDFCkan'

    def entity_call(self, entity, graph=None):
        # Copied from parent class
        dump = graph is None
        if graph is None:
            graph = rdfio.default_graph()
        rdfgenerator = entity.cw_adapt_to(self.adapter)
        rdfgenerator.fill(graph)
        # Also output related entities
        for rtype, role, adapter in (('in_catalog', 'subject', 'RDFCkan'),
                                     ('of_dataset', 'object', 'RDFCkan'),
                                     ('dataset_publisher', 'subject', 'RDFCkan'),
                                     ('dataset_contact_point', 'subject', 'RDFContactPoint')):
            for related in entity.related(rtype, role, entities=True):
                rdfgenerator = related.cw_adapt_to(adapter)
                rdfgenerator.fill(graph)
        # Copied from parent class
        if dump:
            self._dump(graph)


class RDFExportAction(action.Action):

    __regid__ = 'dcat.rdf-export'
    __select__ = one_line_rset() & is_instance('Dataset', 'DataCatalog')
    title = _('RDF export')

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid='dcat.rdf.complete')


class CkanRDFExportAction(action.Action):

    __regid__ = 'dcat.rdf.ckan'
    __select__ = one_line_rset() & is_instance('Dataset', 'DataCatalog')
    title = _('RDF export for CKAN')

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid=self.__regid__)


#
# JSON views
#


class CkanJsonLicenseView(jsonviews.JsonMixIn, EntityView):
    """JSON view outputting data about a license that CKAN can use as ``licenses_group_url``. """

    __regid__ = 'dcat.ckan.json.license'
    __select__ = (EntityView.__select__ & score_entity(lambda c: in_license_scheme(c)))

    def entity_call(self, entity, **kwargs):
        # Fetch Open Definition licenses
        od_path = join(dirname(__file__), 'data', 'od_licenses.json')
        with open(od_path) as f:
            od_licenses = json.load(f)
        url2lic = {}
        title2lic = {}
        for lic_dict in od_licenses.values():
            title2lic[lic_dict['title'].lower()] = lic_dict
            if lic_dict['url']:
                urltuple = urlparse.urlsplit(lic_dict['url'])  # Remove {http://,https://} from URL
                url = urltuple.netloc + urltuple.path
                url2lic[url] = lic_dict
        # Build JSON license
        title = entity.label(language_code=u'en')
        url = entity.exact_match[0].uri if entity.exact_match else entity.cwuri
        urltuple = urlparse.urlsplit(url)
        url0 = urltuple.netloc + urltuple.path
        found = url2lic.get(url0, {})  # Find an Open Definition license with same URL
        if not found:
            for title_variant in (  # Find an Open Definition license with same title
                title,
                title.replace('License', 'Licence'),
                re.sub(r'v(\d)', lambda match: match.group(1), title),  # v1.0 -> 1.0
                re.sub(r'v(\d)', lambda match: match.group(1),
                       title.replace('License', 'Licence')),
            ):
                found = title2lic.get(title_variant.lower(), {})
                if found:
                    break
        license = {'title': title, 'url': url}
        for key, default_value in [
            ('domain_content', False),
            ('domain_data', True),
            ('domain_software', False),
            ('family', u''),
            ('id', entity.cwuri),
            ('maintainer', u''),
            ('od_conformance', u'not reviewed'),
            ('osd_conformance', u'not reviewed'),
            ('status', u'active'),
        ]:
            license[key] = found.get(key, default_value)
        self.w(json_dumps(license, sort_keys=True))


class CkanJsonLicensesView(jsonviews.JsonMixIn, EntityView):
    """JSON view outputting data about a license that CKAN can use as ``licenses_group_url``. """

    __regid__ = 'dcat.ckan.json.licenses'
    __select__ = (EntityView.__select__ &
                  score_entity(lambda cs: cs.scheme_relation and
                               cs.scheme_relation[0].name == u'license_type'))

    def entity_call(self, entity, **kwargs):
        licenses = []
        for concept in self._cw.find('Concept', in_scheme=entity).entities():
            licenses.append(json.loads(concept.view('dcat.ckan.json.license')))
        self.w(json_dumps(licenses, sort_keys=True, indent=2))


#
# HTML views
#

class IDownloadableOutOfContextView(baseviews.OutOfContextView):
    """Out of context view for IDownloadable

    Adapted from cubicweb.web.views.idownloadable.IDownloadableOneLineView
    with content-type and a download icon.

    This is used in `file_distribution` relation contextual component in
    particular.
    """

    __select__ = (baseviews.OutOfContextView.__select__ &
                  adaptable('IDownloadable'))
    download_icon = 'glyphicon glyphicon-download-alt'

    def cell_call(self, row, col):
        entity = self.cw_rset.get_entity(row, col)
        url = xml_escape(entity.absolute_url())
        adapter = entity.cw_adapt_to('IDownloadable')
        name = xml_escape(entity.dc_title())
        durl = xml_escape(adapter.download_url())
        dcontenttype = xml_escape(adapter.download_content_type())
        self.w(u'<a class="{2}" href="{0}" title="{1}"></a> '.format(
               durl, self._cw._('download'), self.download_icon))
        self.w(u'<a href="{0}">{1}</a> '.format(url, name))
        self.w(u'<span class="badge">{0}</span>'.format(dcontenttype))


# File
uicfg.indexview_etype_section['File'] = 'subobject'
pvs.tag_subject_of(('File', 'replaces', '*'), 'hidden')
pvs.tag_object_of(('*', 'replaces', 'File'), 'hidden')
afs.tag_object_of(('*', 'replaces', 'File'), 'main', 'hidden')
abaa.tag_object_of(('*', 'replaces', 'File'), True)

# afs.tag_subject_of(('File', 'resource_of', '*'), 'main', 'hidden')
afs.tag_object_of(('*', 'process_input_file', 'File'), 'main', 'hidden')

pvs.tag_subject_of(('File', 'file_distribution', '*'), 'hidden')


navigation.NextPrevNavigationComponent.context = 'navtop'


class FileIPrevNextAdapter(navigation.IPrevNextAdapter):
    """IPrevNext adapter for file replaced by or replacing another file.
    """

    __select__ = (has_related_entities('replaces', role='subject') |
                  has_related_entities('replaces', role='object'))

    def next_entity(self):
        successors = self.entity.reverse_replaces
        if successors:
            return successors[0]

    def previous_entity(self):
        predecessors = self.entity.replaces
        if predecessors:
            return predecessors[0]


class FileReplacedBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define <New file version>/<Old file version> breadcrumb."""

    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('replaces', role='object'))

    def parent_entity(self):
        return self.entity.reverse_replaces[0]


class DistributionFileBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define Distribution / File breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('file_distribution', role='subject') &
                  # Prevent select ambiguity ad File can be object of both
                  # `implemented_by` and `file_distribution` relations.
                  ~has_related_entities('implemented_by', role='object'))

    def parent_entity(self):
        return self.entity.file_distribution[0]


# Agent
abaa.tag_object_of(('*', 'catalog_publisher', 'Agent'), True)
afs.tag_object_of(('*', 'catalog_publisher', 'Agent'), 'main', 'hidden')
affk.set_field_kwargs('Agent', 'name', widget=fw.TextInput({'size': 82}))
affk.set_field_kwargs('Agent', 'email', widget=fw.TextInput({'size': 82}))
pvs.tag_subject_of(('Agent', 'publisher_type', 'Concept'), 'attributes')
afs.tag_subject_of(('Agent', 'publisher_type', 'Concept'), 'main', 'attributes')


# Concept

def license_scheme_relation_choices(form, field, **kwargs):
    """Choices function for CheckBox widget to select/unselect whether a scheme
    is used to qualify a license_type relationship.
    """
    rtype_eid = form._cw.find('CWRType', name=u'license_type')[0][0]
    return [(u'', unicode(rtype_eid))]


def in_license_scheme(entity):
    """Return 1 if entity is a Concept in a scheme used to describe license
    type.
    """
    if not entity.cw_etype == 'Concept':
        return 0
    if not entity.has_eid():
        return 0
    scheme = entity.in_scheme[0]
    if (scheme.scheme_relation and
            scheme.scheme_relation[0].name == u'license_type'):
        return 1
    return 0


pvs.tag_subject_of(('Concept', 'license_type', 'Concept'), 'hidden')

license_concept_pvs = build_selectable_uicfg(pvs, score_entity(in_license_scheme))
license_concept_pvs.tag_subject_of(('Concept', 'license_type', 'Concept'), 'attributes')


afs.tag_subject_of(('Concept', 'license_type', 'Concept'), 'main', 'hidden')
afs.tag_object_of(('Concept', 'license_type', 'Concept'), 'main', 'hidden')
for rtype in ('publisher_type', 'language', 'dataset_frequency', 'dcat_theme',
              'license', 'media_type', 'distribution_format'):
    afs.tag_object_of(('*', rtype, 'Concept'), 'main', 'hidden')
afs.tag_subject_of(('Concept', 'license', '*'), 'main', 'hidden')

licence_concept_afs = build_selectable_uicfg(afs, score_entity(in_license_scheme))
licence_concept_afs.tag_subject_of(('Concept', 'license_type', 'Concept'), 'main', 'attributes')
afs.tag_object_of(('*', 'theme_taxonomy', 'ConceptScheme'), 'main', 'hidden')
afs.tag_subject_of(('ConceptScheme', 'scheme_relation', '*'), 'main', 'attributes')
affk.tag_subject_of(('ConceptScheme', 'scheme_relation', '*'),
                    {'label': _('describes licenses?'), 'help': u'',
                     'widget': fw.CheckBox, 'choices': license_scheme_relation_choices})


# DataCatalog
pvs.tag_attribute(('DataCatalog', 'title'), 'hidden')
pvs.tag_attribute(('DataCatalog', 'issued'), 'hidden')
afs.tag_attribute(('DataCatalog', 'issued'), 'main', 'hidden')
pvs.tag_attribute(('DataCatalog', 'modified'), 'hidden')
afs.tag_attribute(('DataCatalog', 'modified'), 'main', 'hidden')
pvs.tag_subject_of(('DataCatalog', 'catalog_publisher', '*'), 'hidden')
pvs.tag_subject_of(('DataCatalog', 'license', '*'), 'attributes')
afs.tag_subject_of(('DataCatalog', 'theme_taxonomy', '*'), 'main', 'attributes')
afs.tag_subject_of(('DataCatalog', 'catalog_publisher', '*'), 'main', 'attributes')
abaa.tag_object_of(('*', 'in_catalog', 'DataCatalog'), True)
abaa.tag_subject_of(('DataCatalog', 'theme_taxonomy', '*'), True)
afs.tag_object_of(('*', 'in_catalog', 'DataCatalog'), 'main', 'hidden')
pvdc.tag_attribute(('DataCatalog', 'homepage'), {'vid': 'urlattr'})
for attr in ('title', 'homepage'):
    affk.set_field_kwargs('DataCatalog', attr, widget=fw.TextInput({'size': 82}))
afs.tag_subject_of(('DataCatalog', 'license', 'Concept'), 'main', 'attributes')


class DataCatalogBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define <Agent>/<DataCatalog> breadcrumb."""

    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('catalog_publisher'))

    def parent_entity(self):
        return self.entity.catalog_publisher[0]


# Dataset

def dataset_publisher_choices(form, field, **kwargs):
    """Return form field choices corresponding to all possible publishers for a dataset."""
    # Try to find out if this dataset is linked to a catalog
    linkedto = form.linked_to.get(('in_catalog', 'subject'))
    if linkedto:  # Adding dataset from a catalog view
        catalog = form._cw.entity_from_eid(linkedto[0])
        publishers = catalog.catalog_publisher
        publisher = publishers[0] if publishers else None
        if publisher is not None:
            return [(publisher.dc_title(), unicode(publisher.eid))]
    limit = kwargs.get('limit', 50)
    qs = 'DISTINCT Agent X LIMIT {limit}'.format(limit=limit)
    return [(entity.dc_title(), unicode(entity.eid)) for entity in form._cw.execute(qs).entities()]

pvs.tag_subject_of(('Dataset', 'in_catalog', '*'), 'hidden')
pvs.tag_subject_of(('Dataset', 'dataset_frequency', '*'), 'attributes')
pvs.tag_subject_of(('Dataset', 'dataset_contact_point', '*'), 'attributes')
pvs.tag_subject_of(('Dataset', 'dataset_publisher', '*'), 'attributes')
pvdc.tag_attribute(('Dataset', 'landing_page'), {'vid': 'urlattr'})
afs.tag_subject_of(('Dataset', 'dataset_contact_point', '*'), 'main', 'attributes')
afs.tag_subject_of(('Dataset', 'dataset_publisher', '*'), 'main', 'attributes')
pvs.tag_subject_of(('Dataset', 'dataset_contributors', '*'), 'attributes')
afs.tag_subject_of(('Dataset', 'dataset_contributors', '*'), 'main', 'attributes')
pvs.tag_attribute(('Dataset', 'title'), 'hidden')
pvs.tag_attribute(('Dataset', 'identifier'), 'hidden')
afs.tag_attribute(('Dataset', 'identifier'), 'main', 'hidden')
pvs.tag_attribute(('Dataset', 'issued'), 'hidden')
afs.tag_attribute(('Dataset', 'issued'), 'main', 'hidden')
pvs.tag_attribute(('Dataset', 'modified'), 'hidden')
afs.tag_attribute(('Dataset', 'modified'), 'main', 'hidden')
# Keep dcat_theme hidden from autoform since the RelationWidget will not work
# on creation as constraints on DataCatalog cannot be checked and we prefer
# the user to edit this *inline* from the primary view.
afs.tag_subject_of(('Dataset', 'dcat_theme', '*'), 'main', 'hidden')
pvs.tag_subject_of(('Dataset', 'dcat_theme', '*'), 'attributes')
affk.tag_subject_of(('Dataset', 'dcat_theme', '*'),
                    {'widget': rwdg.RelationFacetWidget(no_creation_form=True)})
afs.tag_subject_of(('Dataset', 'in_catalog', '*'), 'main', 'attributes')
afs.tag_object_of(('*', 'resource_feed_of', 'Dataset'), 'main', 'hidden')
afs.tag_object_of(('*', 'of_dataset', 'Dataset'), 'main', 'hidden')
for attr in ('identifier', 'title', 'keyword', 'landing_page',
             'provenance'):
    affk.set_field_kwargs('Dataset', attr, widget=fw.TextInput({'size': 82}))
afs.tag_subject_of(('Dataset', 'dataset_frequency', 'Concept'),
                   'main', 'attributes')
affk.tag_subject_of(('Dataset', 'dataset_publisher', '*'),
                    {'choices': dataset_publisher_choices})


class DatasetPublisherFacet(cwfacet.RelationFacet):
    __regid__ = 'datacat.dataset_publisher_facet'
    __select__ = cwfacet.RelationFacet.__select__ & is_instance('Dataset')
    rtype = 'dataset_publisher'
    role = 'subject'
    target_attr = 'name'


class DatasetInCatalogFacet(cwfacet.RelationFacet):
    __regid__ = 'datacat.dataset_in_catalog_facet'
    __select__ = cwfacet.RelationFacet.__select__ & is_instance('Dataset')
    rtype = 'in_catalog'
    role = 'subject'
    target_attr = 'title'


class DatasetIssuedFacet(cwfacet.DateRangeFacet):
    """Facet filtering datasets based on their publication date."""
    __regid__ = 'datacat.dataset_issued_facet'
    __select__ = cwfacet.DateRangeFacet.__select__ & is_instance('Dataset')
    rtype = 'issued'


class DatasetModifiedFacet(cwfacet.DateRangeFacet):
    """Facet filtering datasets based on their modification date."""
    __regid__ = 'datacat.dataset_modified_facet'
    __select__ = cwfacet.DateRangeFacet.__select__ & is_instance('Dataset')
    rtype = 'modified'


class DatasetCatalogBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define DataCatalog / Dataset breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('in_catalog', role='subject'))

    def parent_entity(self):
        return self.entity.in_catalog[0]


class DatasetHasDistributionFacet(cwfacet.HasRelationFacet):
    """Facet filtering datasets with a distribution."""
    __regid__ = 'datacat.has_distribution'
    __select__ = cwfacet.HasRelationFacet.__select__ & is_instance('Dataset')
    rtype, role = 'of_dataset', 'object'


# Distribution

class DistributionBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define Dataset / Distribution breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('of_dataset', role='subject'))

    def parent_entity(self):
        return self.entity.of_dataset[0]


pvs.tag_attribute(('Distribution', 'title'), 'hidden')
affk.set_field_kwargs('Distribution', 'title', widget=fw.TextInput({'size': 82}))
afs.tag_object_of(('*', 'file_distribution', 'Distribution'), 'main', 'inlined')
abaa.tag_object_of(('*', 'file_distribution', 'Distribution'), True)
afs.tag_attribute(('Distribution', 'byte_size'), 'main', 'hidden')
afs.tag_attribute(('Distribution', 'download_url'), 'main', 'hidden')
afs.tag_attribute(('Distribution', 'issued'), 'main', 'hidden')
afs.tag_attribute(('Distribution', 'modified'), 'main', 'hidden')
pvs.tag_subject_of(('Distribution', 'license', '*'), 'attributes')
pvs.tag_subject_of(('Distribution', 'media_type', '*'), 'attributes')
pvs.tag_subject_of(('Distribution', 'distribution_format', '*'), 'attributes')
afs.tag_subject_of(('Distribution', 'license', '*'), 'main', 'inlined')
affk.tag_subject_of(
    ('Distribution', 'distribution_format', 'Concept'),
    {'widget': _autocomplete_widget(
        'get_filetypes', placeholder=_('e.g. CSV, PNG, etc.'))}
)
afs.tag_subject_of(('Distribution', 'license', 'Concept'), 'main', 'attributes')


@ajaxcontroller.ajaxfunc(output_type='json')
def get_filetypes(self):
    return search_in_scheme(
        self._cw, 'http://publications.europa.eu/resource/authority/file-type')


@ajaxcontroller.ajaxfunc(output_type='json')
def get_mediattypes(self):
    return search_in_scheme(
        self._cw, 'http://www.iana.org/assignments/media-types/media-types.xml')


pvs.tag_attribute(('Distribution', 'access_url'), 'hidden')
afs.tag_attribute(('Distribution', 'access_url'), 'main', 'hidden')
pvs.tag_attribute(('Distribution', 'download_url'), 'hidden')
afs.tag_attribute(('Distribution', 'download_url'), 'main', 'hidden')
pvs.tag_subject_of(('Distribution', 'of_dataset', '*'), 'hidden')
afs.tag_subject_of(('Distribution', 'of_dataset', '*'), 'main', 'hidden')
pvs.tag_subject_of(('Distribution', 'distribution_format', '*'), 'attributes')
afs.tag_subject_of(('Distribution', 'distribution_format', '*'), 'main', 'attributes')
pvs.tag_subject_of(('Distribution', 'media_type', '*'), 'attributes')
afs.tag_subject_of(('Distribution', 'media_type', '*'), 'main', 'attributes')
pvs.tag_attribute(('Distribution', 'issued'), 'hidden')
afs.tag_attribute(('Distribution', 'issued'), 'main', 'hidden')
pvs.tag_attribute(('Distribution', 'modified'), 'hidden')
afs.tag_attribute(('Distribution', 'modified'), 'main', 'hidden')
pvs.tag_attribute(('Distribution', 'byte_size'), 'hidden')
afs.tag_attribute(('Distribution', 'byte_size'), 'main', 'hidden')
affk.tag_subject_of(
    ('Distribution', 'media_type', 'Concept'),
    {'widget': _autocomplete_widget(
        'get_mediattypes',
        placeholder=_('e.g. text/csv, image/gif, application/octet-stream, etc.'))
     },
)


# XXX does not work except in creation mode.
distr_afs = build_selectable_uicfg(afs, has_related_entities('file_distribution'))
distr_afs.tag_attribute(('File', 'title'), 'main', 'hidden')
distr_afs.tag_attribute(('File', 'description'), 'main', 'hidden')

# ResourceFeed
pvdc.tag_attribute(('ResourceFeed', 'url'), {'vid': 'urlattr'})
pvs.tag_attribute(('ResourceFeed', 'title'), 'hidden')
pvs.tag_subject_of(('ResourceFeed', 'resource_feed_of', '*'), 'hidden')
affk.set_field_kwargs('ResourceFeed', 'title', widget=fw.TextInput({'size': 82}))
affk.set_field_kwargs('ResourceFeed', 'url', widget=fw.TextInput({'size': 82}))
afs.tag_subject_of(('ResourceFeed', 'resource_feed_of', '*'), 'main', 'hidden')
pvs.tag_subject_of(('ResourceFeed', 'media_type', '*'), 'attributes')
afs.tag_subject_of(('ResourceFeed', 'media_type', '*'), 'main', 'attributes')
affk.set_fields_order('ResourceFeed',
                      ('title', 'url', 'media_type', 'validation_script',
                       'transformation_script'))
pvdc.tag_subject_of(('ResourceFeed', 'media_type', '*'), {'order': 2})
pvdc.tag_subject_of(('ResourceFeed', 'validation_script', '*'), {'order': 3})
pvdc.tag_subject_of(('ResourceFeed', 'transformation_script', '*'), {'order': 4})
pvs.tag_object_of(('*', 'process_for_resourcefeed', 'ResourceFeed'), 'hidden')
afs.tag_object_of(('*', 'process_for_resourcefeed', 'ResourceFeed'), 'main', 'hidden')
afs.tag_subject_of(('ResourceFeed', 'resource_feed_source', '*'),
                   'main', 'hidden')
abaa.tag_object_of(('*', 'process_for_resourcefeed', 'ResourceFeed'), False)


class ResourceFeedMediaTypeMasterWidget(fw.LazyRestrictedAutoCompletionWidget):
    """Custom widget that controls the content of "dependent" widgets."""
    needs_js = fw.LazyRestrictedAutoCompletionWidget.needs_js + ('cubes.datacat.js', )

    def render(self, form, field, renderer=None):
        master_domid = field.dom_id(form)
        dependent_info = dict(
            (etype, form.field_by_name(rtype, 'subject').dom_id(form))
            for etype, rtype in [('ValidationScript', 'validation_script'),
                                 ('TransformationScript', 'transformation_script')]
        )
        form._cw.add_onload(js.cw.cubes.datacat.initDependentFormField(
            master_domid, dependent_info))
        return super(ResourceFeedMediaTypeMasterWidget, self).render(form, field, renderer)


def script_choice(etype, form, field):
    """Vocabulary function grouping entities of type `etype` choices by
    media_type.
    """
    req = form._cw
    rset = req.execute('Any X,N,C,LL WHERE X is {0}, X name N, X media_type C,'
                       ' C preferred_label L, L label LL'.format(etype))
    vdict = {}
    for script_eid, script_name, mtype_eid, mtype in rset:
        vdict.setdefault((mtype_eid, mtype), []).append(
            (script_name, str(script_eid)))
    vocab = [(u'', '')]
    for (mtype_eid, mtype), values in vdict.iteritems():
        optgroup_id = '{0}_mediatype_{1}'.format(
            etype.lower(), mtype.replace('/', '-'))
        vocab.append((mtype, None, {'id': optgroup_id}))
        vocab.extend(values)
    return vocab


for etype, rtype in [('ValidationScript', 'validation_script'),
                     ('TransformationScript', 'transformation_script')]:
    afs.tag_subject_of(('ResourceFeed', rtype, '*'), 'main', 'attributes')
    affk.tag_subject_of(('ResourceFeed', rtype, '*'),
                        {'choices': partial(script_choice, etype),
                         'sort': False})
affk.tag_subject_of(
    ('ResourceFeed', 'media_type', 'Concept'),
    {'widget': _autocomplete_widget(
        'get_mediattypes',
        wdgclass=ResourceFeedMediaTypeMasterWidget,
        placeholder=_('e.g. text/csv, image/gif, application/octet-stream, etc.'))
     },
)


class CWSourceBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define ResourceFeed / CWSource breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('resource_feed_source', role='object'))

    def parent_entity(self):
        return self.entity.reverse_resource_feed_source[0]


class ResourceFeedBreadCrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Define Dataset / ResourceFeed breadcrumbs"""
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__ &
                  has_related_entities('resource_feed_of', role='subject'))

    def parent_entity(self):
        """The Dataset"""
        return self.entity.resource_feed_of[0]


class DataProcessInResourceFeedCtxComponent(component.EntityCtxComponent):
    """Display data processes in ResourceFeed primary view"""
    __regid__ = 'datacat.resourcefeed-dataprocess'
    __select__ = (component.EntityCtxComponent.__select__ &
                  is_instance('ResourceFeed') &
                  has_related_entities('process_for_resourcefeed',
                                       role='object'))
    title = _('Data processes')
    context = 'navcontentbottom'

    def render_body(self, w):
        rset = self._cw.execute(
            'Any P,I,S,D WHERE P process_for_resourcefeed X,'
            '                  P process_input_file I,'
            '                  P in_state ST, ST name S,'
            '                  D? process_depends_on P,'
            '                  X eid %(eid)s',
            {'eid': self.entity.eid})
        if rset:
            w(self._cw.view('table', rset=rset))
        rset = self._cw.execute(
            'Any P,I,S,D WHERE P process_for_resourcefeed X,'
            '                  P process_input_file I,'
            '                  P in_state ST, ST name S,'
            '                  P process_depends_on D?,'
            '                  X eid %(eid)s',
            {'eid': self.entity.eid})
        if rset:
            w(self._cw.view('table', rset=rset))



from __future__ import print_function

raise AssertionError('0.7.2 migration is not (completely) implemented')

print('-> creating Eurovoc Source')
source = cnx.create_entity('CWSource', name=u'Eurovoc', url=vocabularies.EUROVOC_URI,
                           type=u'datafeed', parser=u'rdf.skos.zip',  # XXX parser does not exists
                           config=u'synchronize=no\nuse-cwuri-as-url=no')
commit(ask_confirm=False)

sync_schema_props_perms('ResourceFeed')

rename_relation_type('distribution_media_type', 'media_type')

add_relation_definition('ResourceFeed', 'media_type', 'Concept')

mtrql = 'Any X WHERE L label_of X, L label %(label)s'
unspecified_eid = rql(mtrql, {'label': u'application/octet-stream'})[0][0]
for rfeid, df in rql('Any R,F WHERE R is ResourceFeed, R data_format F'):
    rset = rql(mtrql, {'label': df})
    if rset:
        mtype_eid = rset[0][0]
    else:
        print('no matching media type concept for data format {0} of '
              'resource feed #{1}'.format(df, rfeid))
        mtype_eid = unspecified_eid
    rql('SET RF media_type C WHERE RF eid %(rf)s, C eid %(c)s',
        {'rf': rfeid, 'c': mtype_eid})
commit()

drop_attribute('ResourceFeed', 'data_format')

for file_ in rql('Any F WHERE F is File, F cw_source S, NOT S name "system"',
                 ask_confirm=False).entities():
    file_.cw_set(data_sha1hex=file_.compute_sha1hex())
commit(ask_confirm=False)

add_cube('dataprocessing')
sync_schema_props_perms('process_for_resourcefeed')

add_relation_type('output_media_type')



sync_schema_props_perms('dataset_frequency', syncperms=False)
sync_schema_props_perms('license_type', syncperms=False)
sync_schema_props_perms('publisher_type', syncperms=False)
sync_schema_props_perms('scheme_relation', syncperms=False)



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat specific hooks and operations"""

import os
from datetime import datetime

from pytz import utc

from cubicweb.predicates import is_instance, has_related_entities
from cubicweb.server import hook
from cubicweb.server.sources import storages

from cubes.datacat import concept_in


class ServerStartupHook(hook.Hook):
    __regid__ = 'datacat.serverstartup'
    events = ('server_startup', 'server_maintenance')

    def __call__(self):
        bfssdir = os.path.join(self.repo.config.appdatahome, 'bfss')
        if not os.path.exists(bfssdir):
            os.makedirs(bfssdir)
            print 'created', bfssdir
        storage = storages.BytesFileSystemStorage(bfssdir)
        storages.set_attribute_storage(self.repo, 'File', 'data', storage)


class InitDatedEntityHook(hook.Hook):
    """Initialize issued/modified attributes of entities."""
    __regid__ = 'datacat.init-dates'
    __select__ = hook.Hook.__select__ & is_instance('DataCatalog', 'Dataset',
                                                    'Distribution')
    events = ('before_add_entity', )

    def __call__(self):
        now = datetime.now(utc)
        issued = self.entity.cw_edited.setdefault('issued', now)
        self.entity.cw_edited.setdefault('modified', issued)


class UpdateDatedEntityHook(hook.Hook):
    """Update "modified" attribute of entities."""
    __regid__ = 'datacat.update-modified'
    __select__ = hook.Hook.__select__ & is_instance('DataCatalog', 'Dataset',
                                                    'Distribution')
    events = ('before_update_entity', )

    def __call__(self):
        now = datetime.now(utc)
        self.entity.cw_edited.setdefault('modified', now)


class InitDistributionHook(hook.Hook):
    """Initialize "hidden" attributes of Distribution entities."""
    __regid__ = 'datacat.init_distribution'
    __select__ = hook.Hook.__select__ & is_instance('Distribution')
    events = ('before_add_entity', )

    def __call__(self):
        self.entity.cw_edited.setdefault(
            'access_url', self.entity.absolute_url())


class CreateDistributionForResourceFeedHook(hook.Hook):
    """Create a distribution associated with a ResourceFeed upon addition of
    `resource_feed_of` relation.
    """
    __regid__ = 'datacat.resourcefeed_distribution'
    __select__ = hook.Hook.__select__ & hook.match_rtype('resource_feed_of')
    events = ('after_add_relation', )

    def __call__(self):
        resourcefeed = self._cw.entity_from_eid(self.eidfrom)
        # We do not handle "update" of resource_feed_of relation, but this
        # should not occur in practice.
        assert not resourcefeed.resourcefeed_distribution, \
            'ResourceFeed already has a related distribution'
        dataset = self._cw.entity_from_eid(self.eidto)
        self._cw.create_entity('Distribution',
                               title=resourcefeed.title,
                               media_type=resourcefeed.media_type,
                               of_dataset=dataset,
                               reverse_resourcefeed_distribution=resourcefeed)


class UpdateDistributionModificationDateFromFileHook(hook.Hook):
    """Update the modification_date of the Distribution when related to a
    File.
    """
    __regid__ = 'datacat.update-distribution-modification-date-from-file'
    __select__ = hook.Hook.__select__ & has_related_entities('file_distribution')
    events = ('after_update_entity', )

    def __call__(self):
        # XXX spurious implementation, needs review.
        if self.entity.has_eid():
            dist = self.entity.file_distribution[0]
            dist.cw_set(modified=datetime.now(utc))


class DistributionMetadataFromFileHook(hook.Hook):
    """Set Distribution metadata according its related File."""
    __regid__ = 'datacat.set_distribution_metadata_from_file'
    __select__ = hook.Hook.__select__ & hook.match_rtype('file_distribution')
    events = ('after_add_relation', )

    def __call__(self):
        distr = self._cw.entity_from_eid(self.eidto)
        distr_file = self._cw.entity_from_eid(self.eidfrom)
        distr_file.cw_set(title=distr.title, description=distr.description)
        now = datetime.now(utc)
        media_type = concept_in(
            self._cw, label=distr_file.data_format,
            scheme_uri=u'http://www.iana.org/assignments/media-types/media-types.xml')
        distr.cw_set(byte_size=distr_file.size(),
                     media_type=media_type,
                     access_url=distr_file.cw_adapt_to("IDownloadable").download_url(),
                     issued=now,
                     modified=now)


class LinkFeedResourceToDistribution(hook.Hook):
    """Add the `file_distribution` relation to files produced by a
    transformation script when the latter is attached to a ResourceFeed (which
    is in turns related to a Distribution).

    If the Distribution already had a file attached, simply set the `File
    replaces File` relation, which in turns will set the file_distribution as
    per 'datacat.replace-file-distribution' hook.
    """
    __regid__ = 'datacat.link-feedresource-to-distribution'
    __select__ = hook.Hook.__select__ & hook.match_rtype('produced_by')
    events = ('after_add_relation', )

    def __call__(self):
        # TODO: Would be nice to build a selector for this.
        rset = self._cw.execute(
            'Any D WHERE R resourcefeed_distribution D,'
            '            TP process_for_resourcefeed R,'
            '            TP eid %(eid)s',
            {'eid': self.eidto})
        assert rset  # schema
        distr_eid, newfile_eid = rset[0][0], self.eidfrom
        # Make old distribution file replaced by new file.
        rset = self._cw.execute(
            'SET NEWF replaces OLDF WHERE OLDF file_distribution D,'
            '                             D eid %(d)s, NEWF eid %(f)s',
            {'d': distr_eid, 'f': newfile_eid})
        logmsg = 'added file #%d (produced by process #%d) as a new version of distribution %#d'
        # No previous file distribution, setting one.
        if not rset:
            self._cw.execute(
                'SET F file_distribution D WHERE D eid %(d)s, F eid %(f)s',
                {'d': distr_eid, 'f': newfile_eid})
            logmsg = 'added file #%d (produced by process #%d) to distribution %#d'
        self.debug(logmsg, newfile_eid, self.eidto, distr_eid)


class ReplaceFileDistributionHook(hook.Hook):
    """Handle replacement of file_distribution subject upon addition of a
    `File replaces File` relation.
    """
    __regid__ = 'datacat.replace-file-distribution'
    __select__ = hook.Hook.__select__ & hook.match_rtype('replaces')
    events = ('before_add_relation', )

    def __call__(self):
        newf, oldf = self.eidfrom, self.eidto
        self._cw.execute(
            'SET NEWF file_distribution D WHERE OLDF file_distribution D,'
            ' OLDF eid %(o)s, NEWF eid %(n)s',
            {'o': oldf, 'n': newf})


class AddCWSourceForResourceFeedHook(hook.Hook):
    """Add a CWSource for ResourceFeed entities."""
    __regid__ = 'datacat.add_cwsource_for_resourcefeed'
    __select__ = hook.Hook.__select__ & is_instance('ResourceFeed')
    events = ('after_add_entity', )

    def __call__(self):
        url = self.entity.cw_edited['url'].strip()
        rset = self._cw.find('CWSource', url=url)
        if rset:
            source = rset.one()
        else:
            name = url.split('://')[-1][:128]
            source = self._cw.create_entity(
                'CWSource', name=name, url=url, type=u'datafeed',
                config=u'use-cwuri-as-url=no',
                parser=u'datacat.resourcefeed-parser')
        self.entity.cw_set(resource_feed_source=source.eid)


class UpdateCWSourceForResourceFeedHook(hook.Hook):
    """Update ResourceFeed's CWSource when its url changes."""
    __regid__ = 'datacat.update_resourcefeed_cwsource'
    __select__ = hook.Hook.__select__ & is_instance('ResourceFeed')
    events = ('after_update_entity', )

    def __call__(self):
        if 'url' in self.entity.cw_edited:
            url = self.entity.cw_edited['url']
            source = self.entity.resource_feed_source[0]
            source.cw_set(url=url, name=url)


class AddTransformationSequenceHook(hook.Hook):
    """Add a TransformationSequence when a "transformation_script" relation is
    added.
    """
    __regid__ = 'datacat.add-transformation-sequence'
    __select__ = (hook.Hook.__select__
                  & hook.match_rtype('transformation_script'))
    events = ('after_add_relation', )

    def __call__(self):
        tseq = self._cw.create_entity('TransformationSequence')
        self._cw.create_entity('TransformationStep', index=0,
                               step_script=self.eidto, in_sequence=tseq)


class DropTransformationSequenceHook(hook.Hook):
    """Drop TransformationSequence when a "transformation_script" relation is
    deleted.
    """
    __regid__ = 'datacat.drop-transformation-sequence'
    __select__ = (hook.Hook.__select__
                  & hook.match_rtype('transformation_script'))
    events = ('before_delete_relation', )

    def __call__(self):
        tseq = self._cw.entity_from_eid(self.eidto).transformation_sequence
        tseq.cw_delete()



# copyright 2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.

"""cubicweb-datacat postcreate script, executed at instance creation time or when
the cube is added to an existing instance.
"""

from cubes.datacat import cwsource_pull_data, vocabularies


TEST_MODE = config.mode == 'test'

print('-> creating SKOS sources')
for name, url in vocabularies.DCAT_SCHEMES:
    print('   ', name)
    vocabularies.add_source(cnx, name, url)
commit(ask_confirm=False)
if not TEST_MODE:
    print('{0} SKOS sources created, synchronize using "cubicweb-ctl '
          ' sync-schemes <instance>" command'.format(
              len(vocabularies.DCAT_SCHEMES)))
    print('-> importing IANA Media Types')
    vocabularies.media_types_import(cnx)
    commit(ask_confirm=False)

print('-> creating Eurovoc Source')
source = cnx.create_entity('CWSource', name=u'Eurovoc', url=vocabularies.EUROVOC_URI,
                           type=u'datafeed', parser=u'rdf.skos.zip',  # XXX parser does not exists
                           config=u'synchronize=no\nuse-cwuri-as-url=no')
commit(ask_confirm=False)



from logilab.common.shellutils import ProgressBar

from cubes.datacat import cwsource_pull_data, vocabularies


vocabularies.media_types_import(cnx)
commit(ask_confirm=False)

# Add SKOS sources and trigger synchronization
sources = [vocabularies.add_source(cnx, name, url)
           for name, url in vocabularies.DCAT_SCHEMES]
commit(ask_confirm=False)
adms_source_eid = cnx.find('CWSource', url=u'https://joinup.ec.europa.eu/svn/adms/ADMS_v1.00/'
                           u'ADMS_SKOS_v1.00.rdf').one().eid
# Set correct source on concept schemes
sql("UPDATE cw_source_relation SET eid_to={0} WHERE eid_from IN ("
    "SELECT cw_eid FROM cw_conceptscheme WHERE cw_cwuri LIKE 'http://purl.org/adms/%'"
    ")".format(adms_source_eid))
# Set correct source on concepts
sql("UPDATE cw_source_relation SET eid_to={0} WHERE eid_from IN ("
    "SELECT eid_from FROM in_scheme_relation "
    "JOIN cw_conceptscheme ON eid_to = cw_eid WHERE cw_cwuri LIKE 'http://purl.org/adms/%'"
    ")".format(adms_source_eid))
# Set correct source on labels
sql("UPDATE cw_source_relation SET eid_to={0} WHERE eid_from IN ("
    "SELECT cw_label.cw_eid FROM cw_label "
    "JOIN in_scheme_relation ON cw_label.cw_label_of = in_scheme_relation.eid_from "
    "JOIN cw_conceptscheme ON in_scheme_relation.eid_to = cw_conceptscheme.cw_eid "
    "WHERE cw_conceptscheme.cw_cwuri LIKE 'http://purl.org/adms/%'"
    ")".format(adms_source_eid))
commit(ask_confirm=False)
pb = ProgressBar(len(sources), title='synchronizing SKOS sources')
for src in sources:
    cwsource_pull_data(repo, src.through_cw_source[0].eid)
    pb.update()
print('')

# Try to move format attribute into distribution_format relation for distributions
add_relation_type('distribution_media_type')
add_relation_type('distribution_format')
formats = dict(rql('Any LOWER(LL),C WHERE C is Concept, C in_scheme S, S cwuri %(source)s, '
                   'L label_of C, L kind "preferred", L label LL',
                   {'source': 'http://publications.europa.eu/resource/authority/file-type'}))
mediatypes = dict(rql('Any LOWER(LL),C WHERE C is Concept, C in_scheme S, S cwuri %(source)s, '
                      'L label_of C, L kind "preferred", L label LL',
                      {'source': 'http://www.iana.org/assignments/media-types/media-types.xml'}))
rset = rql('Any LOWER(F),D WHERE D is Distribution, D format F')
for format_, dist in rset:
    concept = mediatypes.get(format_)
    if concept is not None:
        rql('Set D distribution_media_type C WHERE D eid %(dist)s, C eid %(concept)s',
            {'dist': dist, 'concept': concept})
    concept = formats.get(format_)
    if concept is not None:  # XXX: value is lost if cannot find matching concept
        rql('Set D distribution_format C WHERE D eid %(dist)s, C eid %(concept)s',
            {'dist': dist, 'concept': concept})
commit()
drop_attribute('Distribution', 'format')

add_relation_definition('DataCatalog', 'language', 'Concept')

# Some relations are now mandatory: fill them with default values
for cat in rql('Any CAT WHERE CAT is DataCatalog, CAT title NULL').entities():
    cat.cw_set(title=u'<not specified>')
for cat in rql('Any CAT WHERE CAT is DataCatalog, CAT description NULL').entities():
    cat.cw_set(description=u'<not specified>')
for ds in rql('Any DS WHERE DS is Dataset, DS description NULL').entities():
    ds.cw_set(description=u'<not specified>')
sync_schema_props_perms()

# New attributes for LicenseDocument
add_attribute('LicenseDocument', 'title')
add_attribute('LicenseDocument', 'uri')

create_entity('Agent', name=u'admin')
rql('SET C catalog_publisher A WHERE A name "admin", NOT C catalog_publisher P')
commit(ask_confirm=False)
sync_schema_props_perms('catalog_publisher')

for etype in ('DataCatalog', 'Dataset', 'Distribution'):
    change_attribute_type(etype, 'issued', 'TZDatetime', ask_confirm=False)
    change_attribute_type(etype, 'modified', 'TZDatetime', ask_confirm=False)
sync_schema_props_perms('issued')
sync_schema_props_perms('modified')

add_cube('relationwidget')



"""cubicweb-datacat unit tests for server objects"""

import tempfile
import urllib2

from cubicweb.devtools.testlib import CubicWebTC

from cubes.datacat import cwsource_pull_data
from utils import create_file, mediatypes_scheme


def transformation_sequence(cnx, script):
    tseq = cnx.create_entity('TransformationSequence')
    cnx.create_entity('TransformationStep', index=0,
                      step_script=script, in_sequence=tseq)
    return tseq


class ResourceFeedParserTC(CubicWebTC):

    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            cat = cnx.create_entity('DataCatalog', title=u'My Catalog', description=u'A catalog',
                                    catalog_publisher=cnx.create_entity('Agent', name=u'Publisher'))
            ds = cnx.create_entity('Dataset', title=u'Test dataset', description=u'A dataset',
                                   in_catalog=cat)
            cnx.commit()
            self.dataset_eid = ds.eid

    def pull_data(self, cwsource):
        return cwsource_pull_data(self.repo, cwsource.eid)

    def test_base(self):
        url = u'file://' + self.datapath('resource.dat')
        with self.admin_access.repo_cnx() as cnx:
            mediatype, = mediatypes_scheme(cnx, u'text/csv')
            resourcefeed = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=mediatype,
                resource_feed_of=self.dataset_eid)
            cnx.commit()
        cwsource = resourcefeed.resource_feed_source[0]
        dist = resourcefeed.resourcefeed_distribution[0]
        stats = self.pull_data(cwsource)
        self.assertEqual(len(stats['created']), 1)
        feid = stats['created'].pop()
        with self.admin_access.repo_cnx() as cnx:
            f = cnx.entity_from_eid(feid)
            self.assertEqual(f.data_name, 'resource.dat')
            self.assertEqual(f.data_format, 'text/csv')
            self.assertEqual(f.file_distribution, [dist])
        # second pull, no update of sha1
        stats = self.pull_data(cwsource)
        self.assertEqual(len(stats['created']), 0)

    def test_with_processes(self):
        url = u'file://' + self.datapath('resource.dat')
        with self.admin_access.repo_cnx() as cnx:
            mediatype, = mediatypes_scheme(cnx, u'text/csv')
            # Validation script.
            vscript_eid = cnx.create_entity(
                'ValidationScript', name=u'validation script',
                media_type=mediatype).eid
            create_file(cnx, 'pass', reverse_implemented_by=vscript_eid)
            # Transformation sequence.
            tscript = cnx.create_entity(
                'TransformationScript',
                name=u'transformation script',
                media_type=mediatype)
            create_file(cnx, open(self.datapath('cat.py')).read(),
                        data_name=u'cat.py',
                        reverse_implemented_by=tscript)
            # Create resource feed.
            resourcefeed = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=mediatype,
                resource_feed_of=self.dataset_eid,
                validation_script=vscript_eid,
                transformation_script=tscript)
            cnx.commit()
        cwsource = resourcefeed.resource_feed_source[0]
        stats = self.pull_data(cwsource)
        assert len(stats['created']) == 1
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.find('DataTransformationProcess',
                            process_for_resourcefeed=resourcefeed.eid)
            process = rset.one()
            self.assertEqual(process.in_state[0].name,
                             'wfs_dataprocess_completed')
            # There should be one result.
            output = cnx.find('File', produced_by=process).one()
            self.assertEqual([r.eid for r in output.file_distribution[0].of_dataset],
                             [self.dataset_eid])

    def test_file_update(self):
        """Update a file between two datafeed pulls"""
        with tempfile.NamedTemporaryFile() as tmpf:
            tmpf.write('coucou')
            tmpf.flush()
            with self.admin_access.repo_cnx() as cnx:
                mtype, = mediatypes_scheme(cnx, u'whatever')
                vscript_eid = cnx.create_entity(
                    'ValidationScript',
                    name=u'validation script',
                    media_type=mtype).eid
                create_file(cnx, 'pass',
                            reverse_implemented_by=vscript_eid)
                # Transformation script and process.
                tscript_eid = cnx.create_entity(
                    'TransformationScript',
                    name=u'transformation script',
                    media_type=mtype).eid
                create_file(cnx, open(self.datapath('reverse.py')).read(),
                            data_name=u'reverse',
                            reverse_implemented_by=tscript_eid)
                resourcefeed = cnx.create_entity(
                    'ResourceFeed', url=u'file://' + tmpf.name,
                    media_type=mtype,
                    validation_script=vscript_eid,
                    transformation_script=tscript_eid,
                    resource_feed_of=self.dataset_eid)
                cnx.commit()
            cwsource = resourcefeed.resource_feed_source[0]
            stats = self.pull_data(cwsource)
            self.assertEqual(len(stats['created']), 1, stats)
            feid = stats['created'].pop()
            expected = {'content': 'uocuoc\n', 'nvalidated': 1, 'nproduced': 1}
            self._check_datafeed_output(feid, vscript_eid,
                                        tscript_eid, expected)
            # Change input file.
            tmpf.write('\nau revoir')
            tmpf.flush()
            stats = self.pull_data(cwsource)
            self.assertEqual(len(stats['created']), 1)
            feid_ = stats['created'].pop()
            # second pull: change input
            expected = {'content': 'riover ua\nuocuoc\n',
                        'nvalidated': 2, 'nproduced': 2}
            self._check_datafeed_output(feid_, vscript_eid,
                                        tscript_eid, expected)
            # Pull one more time, without changing the source.
            # third pull: no change
            stats = self.pull_data(cwsource)
            for k, v in stats.iteritems():
                self.assertFalse(v, '%s: %r' % (k, v))
            # `expected` has not changed.
            self._check_datafeed_output(feid_, vscript_eid,
                                        tscript_eid, expected)

    def _check_datafeed_output(self, feid, vscript_eid, tscript_eid, expected):
        with self.admin_access.repo_cnx() as cnx:
            output = cnx.execute(
                'File X WHERE X produced_by TP, TP process_input_file F, F eid %(f)s',
                {'f': feid}).one()
            if 'content' in expected:
                self.assertEqual(output.read(), expected['content'])
            if 'nvalidated' in expected:
                rset = cnx.execute(
                    'Any X WHERE X validated_by VP, VP validation_script S, S eid %(s)s',
                    {'s': vscript_eid})
                self.assertEqual(len(rset), expected['nvalidated'], rset)
            if 'nproduced' in expected:
                nproduced = expected['nproduced']
                rset = cnx.execute(
                    'Any X WHERE X produced_by TP, TP transformation_sequence SEQ,'
                    ' STEP in_sequence SEQ, STEP step_script S, S eid %(s)s',
                    {'s': tscript_eid})
                self.assertEqual(len(rset), nproduced, rset)

    def test_multiple_resourcefeed_same_url(self):
        url = u'file://' + self.datapath('resource.dat')
        with self.admin_access.repo_cnx() as cnx:
            textplain, textcsv = mediatypes_scheme(
                cnx, u'text/plain', u'text/csv')
            # Transformation scripts.
            tscript_plain = cnx.create_entity(
                'TransformationScript', name=u'text/plain',
                media_type=textplain)
            create_file(cnx, 'print "coucou"',
                        reverse_implemented_by=tscript_plain)
            tscript_csv = cnx.create_entity(
                'TransformationScript', name=u'transformation script',
                media_type=textcsv)
            create_file(cnx, open(self.datapath('cat.py')).read(),
                        data_name=u'cat.py',
                        reverse_implemented_by=tscript_csv)
            # Resource feeds.
            resource1 = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=textcsv,
                resource_feed_of=self.dataset_eid,
                transformation_script=tscript_csv)
            cnx.commit()
            cwsource = resource1.resource_feed_source[0]
            resource2 = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=textplain,
                resource_feed_of=self.dataset_eid,
                transformation_script=tscript_plain)
            cnx.commit()
            assert resource2.resource_feed_source[0] == cwsource
            assert len(cwsource.reverse_resource_feed_source), 2
            self._check_value_error(cwsource)
            tscript_plain.cw_set(media_type=textcsv)
            resource2.cw_set(media_type=textcsv)
            cnx.commit()
            stats = self.pull_data(cwsource)
            self._check_transformations(cnx, stats)

    def _check_value_error(self, cwsource):
        with self.assertRaises(ValueError) as cm:
            self.pull_data(cwsource)
        self.assertIn('MIME types of resource feeds attached',
                      str(cm.exception))

    def _check_transformations(self, cnx, stats):
        # Check both data processes have completed.
        rset = cnx.execute(
            'DataTransformationProcess X WHERE X in_state ST,'
            ' ST name "wfs_dataprocess_completed"')
        self.assertEqual(len(rset), 2)
        # Check resources and produced files.
        self.assertEqual(
            cnx.execute('Any COUNT(X) WHERE X file_distribution D')[0][0], 2)
        self.assertEqual(len(stats['created']), 1, stats)
        feid = stats['created'].pop()
        rset = cnx.execute('File X WHERE X produced_by TP, TP process_input_file F, F eid %(f)s',
                           {'f': feid})
        self.assertEqual(len(rset), 2, rset)

    def test_ftp(self):
        """Check that datafeed monkeypatch for retrieve_url handles ftp protocol"""
        url = u'ftp://user:pwd@does.not.exists'
        with self.admin_access.repo_cnx() as cnx:
            mediatype, = mediatypes_scheme(cnx, u'text/csv')
            resourcefeed = cnx.create_entity(
                'ResourceFeed', url=url,
                media_type=mediatype,
                resource_feed_of=self.dataset_eid)
            cnx.commit()
            cwsource = resourcefeed.resource_feed_source[0]
            dfsource = self.repo.sources_by_eid[cwsource.eid]
            parser = dfsource._get_parser(cnx)
            # Just check we get an urllib2 error, meaning that we went through
            # opening the url.
            self.assertRaises(urllib2.URLError, parser.retrieve_url, url)


if __name__ == '__main__':
    import unittest
    unittest.main()
